#     Copyright (c) 2019-2023 IDEMIA
#     Author: IDEMIA (Philippe Fremy, Florent Oulieres)
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from typing import TYPE_CHECKING, Any, Tuple, cast, Set, Dict, List, Optional
import logging
from collections import defaultdict
import enum

from PySide2.QtWidgets import QMessageBox, QTreeWidgetItem, QApplication, QProgressDialog, \
                            QTreeWidget, QWidget, QAbstractItemView
from PySide2.QtCore import QTimer, Qt

if TYPE_CHECKING:
    pass
from src.gui.ui_git_switch_branch import Ui_GitSwitchBranch
from src.mg_dialog_utils import MgDialogWithRepoList
from src.mg_exec_window import MgExecWindow
from src.mg_utils import extractInt, istrcmp, treeWidgetDeepIterator
from src.mg_repo_info import MgRepoInfo
from src import mg_config as mgc
from src.mg_ensure_info_available import MgEnsureInfoAvailable, RepoInfoFlags
from src.mg_exec_task_item import MgExecTaskGit, MgExecTaskGroup
from src import mg_const

class GroupingBy(enum.Enum):
    NONE = 0
    NAME = 1

groupingLabel = {
    GroupingBy.NONE: 'No grouping',
    GroupingBy.NAME: 'Group by name',
}

class DeleteOrSwitch(enum.Enum):
    DELETE        = enum.auto()
    SWITCH_BRANCH = enum.auto()
    CHECKOUT_TAG  = enum.auto()


class ItemRole(enum.Enum):
    BRANCH_TAG_MIDDLE_NAME = 0
    BRANCH_TAG_END_NAME  = 1
    ITEM_REPOSITORY  = 2


IDX_SWITCH_BRANCH = 0
IDX_SWITCH_TAG = 1

COL_BRANCH_TAG_NAME = 0
COL_NB_REPO = 1
COL_BRANCH_TYPE = 2

logger = logging.getLogger('mg_dialog_git_switch_delete_branch')
dbg = logger.debug

def stripOrigin(branches_remote: List[str]) -> List[str]:
    '''Strip the first part of a remote url (the name of the remote, usually, 'origin')'''
    return list(sorted(set(name.split('/', 1)[1] for name in branches_remote)))


def branchNameIsPresentInRemote(branchName: str, branches_remote: List[str]) -> bool:
    '''Return whether one of the remote branches contains branch name, when ignoring
    the origin part.

    Example:
        branchNameIsPresentInRemote('toto', ['origin/toto']) -> True
        branchNameIsPresentInRemote('toto', ['origin/titi']) -> False
    '''
    return branchName in set(name.split('/', 1)[1] for name in branches_remote)


def remoteBranchesForBranchName(branchName: str, branches_remote: List[str]) -> List[str]:
    '''Return the list of all the remote branches containing the branch <branchName>

    Example:
        remoteBranchesForBranchName('toto', ['origin/toto']) -> ['origin/toto']
        remoteBranchesForBranchName('titi', ['origin/toto']) -> []
        remoteBranchesForBranchName('toto', ['origin/toto', 'origin2/toto']) -> ['origin/toto', 'origin2/toto']
    '''
    return [name
               for name in branches_remote
               if branchName == name.split('/', 1)[1]]


def buildRepoBranchInfo(targetedRepos: List[MgRepoInfo]) -> List[Tuple[str, List[str], List[str]]]:
    repoBranchInfo = [(repo.name,
                       repo.branches_local,
                       # strip the name of the remote from the remote branches
                       stripOrigin(repo.branches_remote))
                             for repo in targetedRepos]
    return repoBranchInfo


def analyseRepoBranchOrTagInfo(repoBranchInfo: List[Tuple[str, List[str], List[str]]]) \
        -> List[ Tuple[str, int, str, List[str]] ]:
    '''Analyse the branches or tag of all the repositories and return an information suitable
    for display about which branch are present in how many repositories, in local or remote form.

    For tag, it will just present information about how many repositories are present with a given tag. Just
    ignore the third part of the tuple which makes no sense for tags.

    Example: analyseRepoBranchOrTagInfo( [
        ('repo1', ['master', 'local_branch'], ['master', 'remote_branch'])
        ('repo2', ['dev', 'local_branch'], ['dev', 'remote_branch'])
    ])
    => [ ('master', 1, 'local and remote', ['repo1']),
         ('dev', 1, 'local and remote', ['repo2']),
         ('local_branch', 2, 'local', ['repo1', 'repo2']),
         ('remote_branch', 2, 'remote', ['repo1', 'repo2']),
       ]

    Example: analyseRepoBranchOrTagInfo( [ ('repo1', ['tag1', 'tag2'], []), ('repo2', ['tag2', 'tag3'], [])])
    => [ ('tag1', 1, 'local', ['repo1'] ),
         ('tag2', 2, 'local', ['repo1', 'repo2']),
         ('tag3', 1, 'local', ['repo2']),
       ]

    '''
    local_br_dict: Dict[str, Set[str]] = defaultdict(set)
    remote_br_dict: Dict[str, Set[str]] = defaultdict(set)
    local_and_remote_br_dict: Dict[str, Set[str]] = defaultdict(set)

    for repo, local_branches, remote_branches in repoBranchInfo:
        set_local_and_remote_branches = set(local_branches) & set(remote_branches)
        set_only_local_branches = set(local_branches) - set_local_and_remote_branches
        set_only_remote_branches = set(remote_branches) - set_local_and_remote_branches
        assert len(set_only_local_branches) + len(set_local_and_remote_branches) == len(local_branches)
        assert len(set_only_remote_branches) + len(set_local_and_remote_branches) == len(remote_branches)

        for local_br in set_only_local_branches:
            local_br_dict[local_br].add(repo)

        for remote_br in set_only_remote_branches:
            remote_br_dict[remote_br].add(repo)

        for local_and_remote_br in set_local_and_remote_branches:
            local_and_remote_br_dict[local_and_remote_br].add(repo)


    # move local branches appearing in other remote branches to local_and_remote
    for name in local_br_dict.keys():
        if len(remote_br_dict[name]) > 0:
            local_and_remote_br_dict[name].update(local_br_dict[name])
            local_br_dict[name] = set()
            local_and_remote_br_dict[name].update(remote_br_dict[name])
            remote_br_dict[name] = set()
            continue

        if len(local_and_remote_br_dict[name]) > 0:
            local_and_remote_br_dict[name].update(local_br_dict[name])
            local_br_dict[name] = set()


    for name in remote_br_dict.keys():
        if name in local_and_remote_br_dict:
            local_and_remote_br_dict[name].update(remote_br_dict[name])
            remote_br_dict[name] = set()


    ret = []
    for name, repos in local_br_dict.items():
        if len(repos) > 0:
            ret.append((name, len(repos), 'local', list(sorted(repos))))
    for name, repos in remote_br_dict.items():
        if len(repos) > 0:
            ret.append((name, len(repos), 'remote', list(sorted(repos))))
    for name, repos in local_and_remote_br_dict.items():
        if len(repos) > 0:
            ret.append((name, len(repos), 'local and remote', list(sorted(repos))))

    return ret


def applyFilterToTree(tree: QTreeWidget, filterText: str) -> None:
    '''Apply a filter to the tree of repositories with the following rules:
    - matching is case insensitive and can match in the middle of a word
    - branch/tag names are searched and hidden if they don't match
    - list of repositories of a branch/tag are shown if the branch/tag is shown
    '''
    filterText = filterText.lower()

    # let's define some utility functions
    def unhideItemParents(item: QTreeWidgetItem) -> None:
        '''Unhide the item and all parent items'''
        it = item
        while it:
            it.setHidden(False)
            it = it.parent()

    def unhideItemChildren(item: QTreeWidgetItem) -> None:
        '''Unhide the item and all its children'''
        item.setHidden(False)
        for idx in range(item.childCount()):
            unhideItemChildren(item.child(idx))

    def unhideItemHierarchy(item: QTreeWidgetItem) -> None:
        unhideItemParents(item)
        unhideItemChildren(item)


    # Unhide everything
    if not filterText:
        for item in treeWidgetDeepIterator(tree):
            item.setHidden(False)
        return

    # Hide everything
    for item in treeWidgetDeepIterator(tree):
        item.setHidden(True)

    # we wander through all items in DFS
    for item in treeWidgetDeepIterator(tree):
        # only end tag/branch names items are relevant for filtering
        if item.data(0, Qt.UserRole) != ItemRole.BRANCH_TAG_END_NAME:
            continue

        data = item.data(0, Qt.ToolTipRole)
        if data and filterText in data.lower():
            unhideItemHierarchy(item)
            continue

        # check other columns
        for i in range(1,3):
            if tree.isColumnHidden(i):
                continue
            if filterText in item.text(i).lower():
                unhideItemHierarchy(item)
                continue



class RepoBranchInfoTreeItem(QTreeWidgetItem):

    def __lt__(self, other: 'QTreeWidgetItem') -> bool:
        col = self.treeWidget().sortColumn()
        if col != 1:
            # regular sorting
            return istrcmp(self.text(col), other.text(col))

        colTextSelf = self.text(col)
        colTextOther = other.text(col)
        if len(colTextSelf) and len(colTextOther) and colTextSelf[0].isdigit() and colTextOther[0].isdigit():
            # natural number sorting if we can
            return extractInt(colTextSelf) < extractInt(colTextOther)

        # regular sort strategy will compare strings and place all number starting strings before others
        return istrcmp(self.text(col), other.text(col))


    def findChildByName(self, col: int, name: str) -> 'Optional[RepoBranchInfoTreeItem]':
        '''Search all direct children for an item with the exact name in the given column'''
        for idx in range(self.childCount()):
            if self.child(idx).text(col) == name:
                return cast(RepoBranchInfoTreeItem, self.child(idx))
        return None


    @staticmethod
    def autoAdjustColumnSize(treeWidget: QTreeWidget) -> None:
        '''Adjust automatically the column size to the largest item'''
        for i in range(treeWidget.columnCount()):
            treeWidget.resizeColumnToContents(i)
        QApplication.processEvents()


def fillBranchTagInfo(repoItemInfo: List[Tuple[str, int, str, List[str]]],
                      treeWidget: QTreeWidget,
                      grouping: GroupingBy) -> None:
    '''Fills the treeWidget with items and a nested structure according to grouping
    and repoItemInfo. Works for both tags and branch repoItemInfo

    See analyseRepoBranchOrTagInfo() for the details of the structure.
    '''
    treeWidget.clear()
    if grouping == GroupingBy.NONE:
        for name, count, infoLocalRemote, branchList in repoItemInfo:
            item = RepoBranchInfoTreeItem([name, '%d' % count, infoLocalRemote])
            item.setData(0, Qt.ToolTipRole, name)
            item.setData(1, Qt.ToolTipRole, 'Present in:\n' + '\n'.join(branchList))
            item.setData(0, Qt.UserRole, ItemRole.BRANCH_TAG_END_NAME)
            for branch in branchList:
                childItem = QTreeWidgetItem(['', '    ' + branch])
                childItem.setData(0, Qt.UserRole, ItemRole.ITEM_REPOSITORY)
                item.addChild(childItem)
            item.setExpanded(False)
            treeWidget.addTopLevelItem(item)
        treeWidget.sortByColumn(0, Qt.AscendingOrder)
        return

    if grouping == GroupingBy.NAME:
        repoItemInfo.sort()

        def propagateItemContent(parentItem: RepoBranchInfoTreeItem,
                                 splittedName: List[str],
                                 fullName: str,
                                 count: int,
                                 infoLocalRemote: str,
                                 branchList: List[str]) -> None:
            '''Continue to create item to represent the full splitted branch/tag name.

            A parent item already exists, the role of this function is to create one child item for each
            part of splittedName. When the name is exhausted, items with the list of repositories are added.
            '''
            assert len(splittedName)
            # if we don't have and item with our splitted name part in the childs of our parent, create it
            if parentItem.childCount() == 0 or parentItem.findChildByName(0, splittedName[0]) is None:
                childItem: QTreeWidgetItem = RepoBranchInfoTreeItem([splittedName[0]])
                # Note that with the information available so far, we set it as middle-name. But if it turns
                # out that the name is the final part, this will be overridden with BRANCH_TAG_END_NAME
                childItem.setData(0, Qt.UserRole, ItemRole.BRANCH_TAG_MIDDLE_NAME)
                parentItem.addChild(childItem)
                parentItem.setExpanded(True)

            item = parentItem.findChildByName(0, splittedName[0])
            assert item is not None
            if len(splittedName) == 1:
                # we just filled the end of the item name, also fill other columns
                item.setText(1, str(count))
                item.setText(2, infoLocalRemote)
                item.setData(0, Qt.ToolTipRole, fullName)
                item.setData(1, Qt.ToolTipRole, 'Present in:\n' + '\n'.join(branchList))
                item.setData(0, Qt.UserRole, ItemRole.BRANCH_TAG_END_NAME)
                for branch in branchList:
                    childItem = QTreeWidgetItem(['', '    ' + branch])
                    childItem.setData(0, Qt.UserRole, ItemRole.ITEM_REPOSITORY)
                    item.addChild(childItem)
                item.setExpanded(False)
            else:
                # more items to fill
                propagateItemContent(item, splittedName[1:], fullName, count, infoLocalRemote, branchList)


        lastTopLevel: Optional[RepoBranchInfoTreeItem] = None
        for name, count, infoLocalRemote, branchList in repoItemInfo:
            splittedName = name.split('/')
            if lastTopLevel is None or splittedName[0] != lastTopLevel.text(0):
                # create a new top-level item
                item = RepoBranchInfoTreeItem(treeWidget, [splittedName[0]])
            else:
                item = lastTopLevel
            if len(splittedName) > 1:
                item.setData(0, Qt.UserRole, ItemRole.BRANCH_TAG_MIDDLE_NAME)
                propagateItemContent(item, splittedName[1:], name, count, infoLocalRemote, branchList)
            else:
                item.setText(1, str(count))
                item.setText(2, infoLocalRemote)
                item.setData(0, Qt.ToolTipRole, name)
                item.setData(1, Qt.ToolTipRole, 'Present in:\n' + '\n'.join(branchList))
                item.setData(0, Qt.UserRole, ItemRole.BRANCH_TAG_END_NAME)
                for branch in branchList:
                    childItem = QTreeWidgetItem(['', '    ' + branch])
                    childItem.setData(0, Qt.UserRole, ItemRole.ITEM_REPOSITORY)
                    item.addChild(childItem)
                item.setExpanded(False)

            lastTopLevel = item

    return


class MgDialogGitSwitchDeleteBranch(MgDialogWithRepoList):
    ui: Ui_GitSwitchBranch
    progressDialog: QProgressDialog

    def __init__(self, parent: QWidget,
                 deleteOrSwitch: DeleteOrSwitch,
                 selectedRepos:
                 List[MgRepoInfo],
                 allRepos: List[MgRepoInfo]
                 ) -> None:
        super().__init__(parent, Ui_GitSwitchBranch, selectedRepos, allRepos)

        self.repoNamesMissingBranchOrTagInfo: Set[str] = set()
        self.deleteOrSwitch = deleteOrSwitch

        self.ui.pushButtonGrouping.clicked.connect(self.slotChangeGrouping)
        self.sigRepoListAdjusted.connect(self.ensureBranchTagInfoAvailable)

        self.ui.pushButtonGrouping.setFixedSize(self.ui.pushButtonGrouping.sizeHint())
        self.grouping = GroupingBy.NONE
        self.slotChangeGrouping()

        self.ui.treeWidgetBranches.clear()
        self.ui.treeWidgetBranches.setSortingEnabled(True)
        self.ui.treeWidgetBranches.sortByColumn(0, Qt.AscendingOrder)
        self.ui.treeWidgetBranches.itemSelectionChanged.connect(self.slotItemSelectionChanged)

        self.ui.lineEditBranchFilter.setPlaceholderText('Filter the list of %s by typing here' % ('branches' if self.isBranchDialog() else 'tags'))
        self.ui.lineEditBranchFilter.setClearButtonEnabled(True)
        self.ui.lineEditBranchFilter.textEdited.connect(self.slotApplyFilter)
        self.ui.lineEditBranchTagName.setPlaceholderText('Choose %s from list below or type it here' % ('branch' if self.isBranchDialog() else 'tag'))

        if deleteOrSwitch == DeleteOrSwitch.DELETE:
            self.setWindowTitle('Git Delete Branch')
            self.ui.labelBranchOrTag.setText('Git delete branch')
            self.ui.labelBranchOrTag.setVisible(True)
            self.ui.checkBoxDefaultForNotExist.setVisible(False)
            self.ui.checkBoxDeleteLocalBranch.setVisible(True)
            self.ui.checkBoxDeleteRemoteBranch.setVisible(True)
            self.ui.checkBoxDefaultForNotExist.setVisible(False)
            self.ui.treeWidgetBranches.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
            self.ui.treeWidgetBranches.headerItem().setText(0, 'Branches')
        elif deleteOrSwitch == DeleteOrSwitch.SWITCH_BRANCH:
            self.setWindowTitle('Git Switch Branch')
            self.ui.labelBranchOrTag.setText('Choose branch')
            self.ui.groupBoxBranchOrTagSelection.setTitle('Branch selection')
            self.ui.labelBranchOrTag.setVisible(True)
            self.ui.checkBoxDeleteRemoteBranch.setVisible(False)
            self.ui.checkBoxDeleteLocalBranch.setVisible(False)
            self.ui.checkBoxDefaultForNotExist.setVisible(True)
            self.ui.treeWidgetBranches.setColumnHidden(COL_BRANCH_TYPE, False)
            self.ui.treeWidgetBranches.headerItem().setText(0, 'Branches')
        else:
            assert deleteOrSwitch == DeleteOrSwitch.CHECKOUT_TAG
            self.setWindowTitle('Git Checkout Tag')
            self.ui.labelBranchOrTag.setText('Choose tag')
            self.ui.groupBoxBranchOrTagSelection.setTitle('Tag selection')
            self.ui.labelBranchOrTag.setVisible(True)
            self.ui.checkBoxDeleteRemoteBranch.setVisible(False)
            self.ui.checkBoxDeleteLocalBranch.setVisible(False)
            self.ui.checkBoxDefaultForNotExist.setVisible(False)
            self.ui.treeWidgetBranches.setColumnHidden(COL_BRANCH_TYPE, True)
            self.ui.treeWidgetBranches.headerItem().setText(0, 'Tags')


    def exec_(self) -> Any:
        QTimer.singleShot(0, self.ensureBranchTagInfoAvailable)
        return super().exec_()


    def slotItemSelectionChanged(self) -> None:
        '''The selection has changed. In delete mode, multiple items may be selected.
        In switch mode, only one branch/tag may be selected so this will contain only one item'''
        items = self.ui.treeWidgetBranches.selectedItems()
        if len(items) == 0:
            # nothing selected, difficult to believe
            return

        branchNames = [self.resolveBranchName(item) for item in items]
        branchNamesExcludingNone = [name for name in branchNames if name]
        self.ui.lineEditBranchTagName.setText( '  '.join(branchNamesExcludingNone) )


    def resolveBranchName(self, item: QTreeWidgetItem) -> str:
        '''Return the full branch name, when branches are grouped in a hierarchy:

        + feat/                 -> if selected, returns ''
            + new_feat_1        -> if selected, returns feat/new_feat_1
            + new_feat_2        -> if selected, returns feat/new_feat_2

        '''
        if item.data(0, Qt.UserRole) == ItemRole.ITEM_REPOSITORY:
            # when pointing to a repository, the branch/tag name ends in his parent
            item = item.parent()

        if item.data(0, Qt.UserRole) == ItemRole.BRANCH_TAG_MIDDLE_NAME:
            # for the middle-name, we don't want to report any text
            return ''

        assert item.data(0, Qt.UserRole) == ItemRole.BRANCH_TAG_END_NAME

        # item has parent. Collect the name of all parents + name of the item to form
        # the full branch/tag name.
        visitItem = item
        fullBranchTagName = visitItem.text(0)
        while visitItem.parent() is not None:
            visitItem = visitItem.parent()
            fullBranchTagName = '%s/%s' % (visitItem.text(0), fullBranchTagName)
        return fullBranchTagName


    def ensureBranchTagInfoAvailable(self) -> None:
        '''Collect all branches or tags from all selected repositories. If some repositories
        do not have the branch or tag information set yet, trigger the information reading on the repository.
        If the operation takes too long, show a progress dialog
        '''
        QApplication.processEvents()
        targetedRepos = self.getTargetedRepoList()
        self.ensureInfoAvailable = MgEnsureInfoAvailable(self, targetedRepos, showProgressDialog=True)

        if self.deleteOrSwitch in (DeleteOrSwitch.SWITCH_BRANCH, DeleteOrSwitch.DELETE):
            repoInfo = RepoInfoFlags.ALL_BRANCHES
        else:
            repoInfo = RepoInfoFlags.ALL_TAGS

        self.ensureInfoAvailable.ensureInfoAvailable(repoInfo, blocking=True)
        self.fillTreeWidgetBranchTagSelection()


    def slotChangeGrouping(self) -> None:
        '''Change the grouping option used to display all the branches'''
        try:
            self.grouping = GroupingBy(self.grouping.value+1)
        except ValueError:
            self.grouping = GroupingBy.NONE
        try:
            nextGrouping = GroupingBy(self.grouping.value+1)
        except ValueError:
            nextGrouping = GroupingBy.NONE

        self.ui.pushButtonGrouping.setText(groupingLabel[nextGrouping])
        self.fillTreeWidgetBranchTagSelection()


    def fillTreeWidgetBranchTagSelection(self) -> None:
        '''Fills the tree widget dedicated to displaying all possible branches'''
        targetedRepos = self.getTargetedRepoList()

        repoItemInfo: List[ Tuple[str, int, str, List[str]] ]
        repoBranchTagInfo: List[ Tuple[str, List[str], List[str]]]
        if self.isBranchDialog():
            repoBranchTagInfo = buildRepoBranchInfo(targetedRepos)
        else:
            repoBranchTagInfo = [(repo.name, repo.all_tags, [])
                           for repo in targetedRepos]

        repoItemInfo = analyseRepoBranchOrTagInfo(repoBranchTagInfo)
        self.ui.treeWidgetBranches.clear()
        fillBranchTagInfo(repoItemInfo, self.ui.treeWidgetBranches, self.grouping)
        self.ui.treeWidgetBranches.sortByColumn(0, Qt.AscendingOrder)
        RepoBranchInfoTreeItem.autoAdjustColumnSize(self.ui.treeWidgetBranches)
        self.slotApplyFilter()



    def slotApplyFilter(self) -> None:
        '''Called when the user modifies text of the branch line edit. Trigger filtering
        the content of the items'''
        filterText = self.ui.lineEditBranchFilter.text().lower()
        applyFilterToTree(self.ui.treeWidgetBranches, filterText)

    def getTargetedBranchTag(self) -> str:
        '''Get the branch selected in either local or remote tree'''
        if self.deleteOrSwitch == DeleteOrSwitch.DELETE:
            raise ValueError('Do not use this method in DELETE branch mode')
        return self.ui.lineEditBranchTagName.text()


    def getDeleteTargetedBranches(self) -> List[str]:
        '''Get the branch selected in either local or remote tree'''
        if self.deleteOrSwitch != DeleteOrSwitch.DELETE:
            raise ValueError('Do not use this method when not in DELETE branch mode')
        branchesText =  self.ui.lineEditBranchTagName.text()
        branchesWithEmpty = branchesText.split(' ')
        branches = [br for br in branchesWithEmpty if br]
        return branches


    def isBranchDialog(self) -> bool:
        '''Return True if dialog is about branch (switching, deleting)'''
        return self.deleteOrSwitch in (DeleteOrSwitch.SWITCH_BRANCH, DeleteOrSwitch.DELETE)


    def checkAcceptDeleteBranch(self, targetBranch: str) -> bool:
        '''Check if delete branch is consistent and return True if this is the case.

        If inconsistent, asks the user what he wants to do and return True/False depending on his answer'''
        repoBranchInfo = buildRepoBranchInfo(self.getTargetedRepoList())

        if not self.ui.checkBoxDeleteRemoteBranch.isChecked() and not self.ui.checkBoxDeleteLocalBranch.isChecked():
            msg = f'You must check at least one of  <i>Delete local branch</i> and <i>Delete remote branch</i>.'
            msg += '<p>What do you want to do ?<p>'
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle('Must choose local or remote branch to delete ?')
            msgBox.setTextFormat(Qt.RichText)
            msgBox.setText(msg)

            delLocalButton = msgBox.addButton('Delete local branch', QMessageBox.AcceptRole)
            delRemoteButton = msgBox.addButton('Delete remote branch', QMessageBox.AcceptRole)
            delBothButton = msgBox.addButton('Delete local and remote branch', QMessageBox.AcceptRole)
            cancelButton = msgBox.addButton(QMessageBox.Cancel)
            msgBox.setDefaultButton(cancelButton)
            msgBox.exec()
            buttonSelected = msgBox.clickedButton()

            if buttonSelected == cancelButton:
                # user changed its mind ...
                return False

            if buttonSelected in (delLocalButton, delBothButton):
                self.ui.checkBoxDeleteLocalBranch.setChecked(True)

            if buttonSelected in (delRemoteButton, delBothButton):
                self.ui.checkBoxDeleteRemoteBranch.setChecked(True)


        localRepoWithTargetedBranch = [repo
                                       for repo, repoLocalBranches, repoRemoteBranches in repoBranchInfo
                                       if targetBranch in repoLocalBranches
                                       ]

        remoteRepoWithTargetedBranch = [repo
                                        for repo, repoLocalBranches, repoRemoteBranches in repoBranchInfo
                                        if targetBranch in repoRemoteBranches
                                        ]

        if len(localRepoWithTargetedBranch) + len(remoteRepoWithTargetedBranch) == 0:
            # this branch name does not exist in any repos!
            QMessageBox.warning(self, "Invalid branch name",
                                'No repository exists with the branch: %s' % targetBranch)
            return False

        # branch exists only remotely but remote check box is not checked
        if len(localRepoWithTargetedBranch) == 0 and len(remoteRepoWithTargetedBranch) > 0 \
                and not self.ui.checkBoxDeleteRemoteBranch.isChecked():
            msg = f'You want to delete branch {targetBranch}, which exists only remotely in the repositories '
            msg += 'but you did not check the <i>Delete remote branch</i>'
            msg += '<p>What do you want to do ?<p>'
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle('Delete remote branch ?')
            msgBox.setTextFormat(Qt.RichText)
            msgBox.setText(msg)

            continueButton = msgBox.addButton('Continue', QMessageBox.AcceptRole)
            delButton = msgBox.addButton('Delete remote branch', QMessageBox.AcceptRole)
            cancelButton = msgBox.addButton(QMessageBox.Cancel)
            msgBox.setDefaultButton(delButton)
            msgBox.exec()
            buttonSelected = msgBox.clickedButton()

            if buttonSelected == cancelButton:
                # user changed its mind ...
                return False

            if buttonSelected == delButton:
                self.ui.checkBoxDeleteRemoteBranch.setChecked(True)

        # branch exists only locally but local check box is not checked
        if len(localRepoWithTargetedBranch) > 0 and len(remoteRepoWithTargetedBranch) == 0 \
                and not self.ui.checkBoxDeleteLocalBranch.isChecked():
            msg = f'You want to delete branch {targetBranch}, which exists only locally in the repositories '
            msg += 'but you did not check the <i>Delete local branch</i>'
            msg += '<p>What do you want to do ?<p>'
            msgBox = QMessageBox(self)
            msgBox.setWindowTitle('Delete local branch ?')
            msgBox.setTextFormat(Qt.RichText)
            msgBox.setText(msg)

            continueButton = msgBox.addButton('Continue', QMessageBox.AcceptRole)
            delButton = msgBox.addButton('Delete local branch', QMessageBox.AcceptRole)
            cancelButton = msgBox.addButton(QMessageBox.Cancel)
            msgBox.setDefaultButton(delButton)
            msgBox.exec()
            buttonSelected = msgBox.clickedButton()

            if buttonSelected == cancelButton:
                # user changed its mind ...
                return False

            if buttonSelected == delButton:
                self.ui.checkBoxDeleteLocalBranch.setChecked(True)

        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_GIT_BRANCH_HISTORY, targetBranch)
        return True


    def checkAcceptSwitchBranch(self) -> bool:
        '''Check if switching to branch or tag is coherent'''
        repoBranchInfo = buildRepoBranchInfo(self.getTargetedRepoList())

        repoWithTargetedBranch = [repo
                                  for repo, repoLocalBranches, repoRemoteBranches in repoBranchInfo
                                  if self.getTargetedBranchTag() in repoLocalBranches
                                  or self.getTargetedBranchTag() in repoRemoteBranches
                                  ]

        if len(repoWithTargetedBranch) == 0:
            # this branch name does not exist in any repos!
            QMessageBox.warning(self, "Invalid branch name",
                                'No repository exists with the branch: %s' % self.getTargetedBranchTag())
            return False

        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_GIT_BRANCH_HISTORY, self.getTargetedBranchTag())
        return True


    def checkAcceptCheckoutTag(self) -> bool:
        # switch to tag
        all_repo_tags: Set[str] = set()
        for repo in self.getTargetedRepoList():
            all_repo_tags.update(repo.all_tags)

        if not self.getTargetedBranchTag() in all_repo_tags:
            QMessageBox.warning(self, "Invalid tag name",
                                'No repository exists with the tag: %s' % self.getTargetedBranchTag())
            return False

        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_TAG_HISTORY, self.getTargetedBranchTag())
        return True


    def accept(self) -> None:
        if ((self.deleteOrSwitch in (DeleteOrSwitch.SWITCH_BRANCH, DeleteOrSwitch.CHECKOUT_TAG) and self.getTargetedBranchTag() == '')
            or (self.deleteOrSwitch == DeleteOrSwitch.DELETE and self.getDeleteTargetedBranches() == [])):
            # no branch/tag is actually selected!
            branchOrTag = 'branch' if self.isBranchDialog() else 'tag'
            QMessageBox.warning(self, "No %s selected" % branchOrTag,
                                "You did not select a %s to checkout!" % branchOrTag)
            return

        if self.deleteOrSwitch == DeleteOrSwitch.DELETE:
            for targetBranch in self.getDeleteTargetedBranches():
                if not self.checkAcceptDeleteBranch(targetBranch):
                    return

        elif self.deleteOrSwitch == DeleteOrSwitch.SWITCH_BRANCH:
            if not self.checkAcceptSwitchBranch():
                return

        elif self.deleteOrSwitch == DeleteOrSwitch.CHECKOUT_TAG:
            if not self.checkAcceptCheckoutTag():
                return

        else:
            raise ValueError('Should not be reached!')

        mgc.get_config_instance().save()
        super().accept()


def runDialogGitSwitchDelete(parent: QWidget, deleteOrSwitch: DeleteOrSwitch, selectedRepos: List[MgRepoInfo], allRepos: List[MgRepoInfo]) -> None:
    '''Run a dialog to switch or delete a branch'''
    dbg('runDialogGitSwitchDelete')

    dialog = MgDialogGitSwitchDeleteBranch(parent, deleteOrSwitch, selectedRepos, allRepos)
    result = dialog.exec_()
    if not result:
        # command execution canceled
        return

    if deleteOrSwitch == DeleteOrSwitch.DELETE:
        doGitDeleteBranch(parent, dialog)
    elif deleteOrSwitch in (DeleteOrSwitch.SWITCH_BRANCH, DeleteOrSwitch.CHECKOUT_TAG):
        if not doGitSwitchBranchTag(parent, dialog):
            # re-run the dialog
            runDialogGitSwitchDelete(parent, deleteOrSwitch, selectedRepos, allRepos)
    else:
        raise ValueError('No such value: %s' % deleteOrSwitch)


def doGitDeleteBranch(parent: QWidget, dialog: MgDialogGitSwitchDeleteBranch) -> None:

    # use the same history as git create branch
    descGitDelete = 'Git Delete branch'

    descRepoCmdBranch = []
    deleteBranchesName = dialog.getDeleteTargetedBranches()
    for repo in dialog.getTargetedRepoList():
        gitCmds = []

        for branchName in deleteBranchesName:

            gitDeleteRemoteBranch = ['push', 'origin', '--delete', branchName]
            gitDeleteLocalBranch = ['branch', '-d', branchName]

            ### Delete remote first, so that we can delete local branch even if not fully merged
            #   into remote.
            if branchName in [name[name.index('/') + 1:] for name in
                                 repo.branches_remote] and dialog.ui.checkBoxDeleteRemoteBranch.isChecked():
                # delete remotely
                gitCmds.append(gitDeleteRemoteBranch)

            if branchName in repo.branches_local and dialog.ui.checkBoxDeleteLocalBranch.isChecked():
                # delete locally
                gitCmds.append(gitDeleteLocalBranch)

        if len(gitCmds) > 0:
            descRepoCmdBranch.append((descGitDelete, repo, gitCmds))

    # show window for executing git
    if len(descRepoCmdBranch):
        gitExecWindow = MgExecWindow(parent)
        gitExecWindow.execEachRepoWithHisSeqOfGitCommand(descGitDelete, descRepoCmdBranch)
    else:
        QMessageBox.warning(parent, 'No repository',
                            'No repository found for branch names: {}.'.format(', '.join(deleteBranchesName)))


def doGitSwitchBranchTag(parent: QWidget, dialog: MgDialogGitSwitchDeleteBranch) -> bool:
    # switch branch/tag
    branchTagName = dialog.getTargetedBranchTag()

    gitCheckoutBranch = [['checkout', branchTagName, '--']]
    gitCheckoutBranchInt = [['checkout', 'int', '--']]
    descCheckout = 'Checkouting %s %s ' % ('branch' if dialog.isBranchDialog() else 'tag', branchTagName)
    descCheckoutInt = 'Checkouting backup branch int'
    taskGroupsCmdBranchTag = []
    taskGroupsCmdInt = []

    if dialog.isBranchDialog():
        # checkouting a branch
        for repo in dialog.getTargetedRepoList():
            if branchTagName in repo.branches_local:
                # local branch found
                taskGroupsCmdBranchTag.append(
                    MgExecTaskGroup(descCheckout, repo,
                                                        [MgExecTaskGit(descCheckout, repo, cmdLine)
                                                         for cmdLine in gitCheckoutBranch]))

            elif branchNameIsPresentInRemote(branchTagName, repo.branches_remote):
                # remote branch found!

                # present in multiple origins ?
                remoteBranches = remoteBranchesForBranchName(branchTagName, repo.branches_remote)
                if len(remoteBranches) == 1:
                    # ok, only one origin, simple command:
                    taskGroupsCmdBranchTag.append(
                        MgExecTaskGroup(descCheckout, repo,
                                        [MgExecTaskGit(descCheckout, repo, cmdLine)
                                         for cmdLine in gitCheckoutBranch]))

                else:
                    # multiple origins, ask the user
                    msg1 = f'Multiple remote branches match the name "{branchTagName}" in repository "{repo.name}"\n'
                    msg1 += 'Please select the one you want.'
                    msgBox = QMessageBox(parent)
                    msgBox.setText(msg1)
                    msgBox.setIcon(QMessageBox.Warning)
                    msgBox.setWindowTitle('Multiple remote branch with same name')
                    buttons = []
                    for branch in remoteBranches:
                        buttons.append(msgBox.addButton(branch, QMessageBox.ActionRole))
                    msgBox.setStandardButtons(QMessageBox.Abort)

                    stdButtonClicked = msgBox.exec_()
                    if stdButtonClicked == QMessageBox.Abort:
                        # show the dialog again
                        return False

                    clickedButton = msgBox.clickedButton()
                    remoteBranch = clickedButton.text()
                    taskGroupsCmdBranchTag.append(
                        MgExecTaskGroup(f'Checkouting branch {remoteBranch}', repo,
                                        [MgExecTaskGit(f'Checkouting branch {remoteBranch}', repo,
                                                ['checkout', '--track', remoteBranch, '--'] )]
                                        )
                    )


            elif dialog.ui.checkBoxDefaultForNotExist.isChecked():
                # branch not found and default branch specified
                taskGroupsCmdInt.append(
                    MgExecTaskGroup(descCheckoutInt, repo,
                                    [MgExecTaskGit(descCheckoutInt, repo, cmdLine)
                                     for cmdLine in gitCheckoutBranchInt]))

            # else:
            #    do Nothing
    else:
        # checkouting a tag
        # TODO: when tag does not exist and we know it in advance, display a message instead of trying to checkout it
        for repo in dialog.getTargetedRepoList():
            taskGroupsCmdBranchTag.append(
                MgExecTaskGroup(descCheckout, repo, [
                    # we force git to checkout another version first, so that it reflects the tag name in git status
                    MgExecTaskGit(f'{descCheckout} (step 1)', repo, ['checkout', f'{branchTagName}~1', '--'], ignore_failure=True),
                    MgExecTaskGit(f'{descCheckout} (step 2)', repo, ['checkout', branchTagName, '--']),
                ]))

    # show window for executing git
    gitExecWindow = MgExecWindow(parent)
    if len(taskGroupsCmdBranchTag):
        gitExecWindow.execTaskGroups(descCheckout, taskGroupsCmdBranchTag)
    if len(taskGroupsCmdInt):
        gitExecWindow.execTaskGroups(descCheckoutInt, taskGroupsCmdInt)


    return True

# TODO: group branch by number of repositories
# TODO: show names of the repositories when grouping



#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import cast, Match, Any, Iterable, Sequence, Union, List, Iterator
import html, os, pathlib, re, shutil, stat, logging

from PySide2.QtWidgets import QTreeWidget, QTreeWidgetItem

from src.mg_const import MSG_BIG_DIFF, GIT_AUTH_FAILURE_MARKER

logger = logging.getLogger('mg_utils')
dbg = logger.debug
warn = logger.warning

HTML_HEADER='''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li { white-space: pre-wrap;font-family:'Courier New'; font-size:8pt; font-weight:600; 
        font-style:normal; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; 
        -qt-block-indent:0; text-indent:0px; }
</style></head>
<body style="font-family:'Courier New'; font-size:8pt; font-weight:400; font-style:normal">'''
HTML_FOOTER='</body></html>'
TAG_BLUE_OPEN   = '<span style="background-color:#aaffff;">'
TAG_YELLOW_OPEN = '<span style="background-color:#ffff7f ">'
TAG_GREEN_OPEN  = '<span style="background-color:#ccffcc;">'
TAG_RED_OPEN    = '<span style="background-color:#ffdddd;">'
TAG_GREY_OPEN   = '<span style="background-color:#f5f5f5;">'
TAG_BLUE_CLOSE  = '</span>'
TAG_GREEN_CLOSE = '</span>'
TAG_YELLOW_CLOSE= '</span>'
TAG_RED_CLOSE   = '</span>'
TAG_GREY_CLOSE  = '</span>'
# TAG_YELLOW_OPEN = '<span style="color:#800000;background-color:#ffff7f ">'
TAG_NL = '<br/>'
re_highlight = re.compile(r'^(---|\+\+\+|diff|index).*')
re_hunk_header = re.compile('^(@@.+?@@)(.*)')
re_line_added = re.compile(r'^\+.*')
re_line_removed = re.compile(r'^-.*')
re_stat_line = re.compile(r'( .* \|.* \d+ )(\+*)(-*)(.*)')


def htmlize_diff(s: str, maxLines: int = -1) -> str:
    '''Transform a diff multi-line string into a html colorized
    string suitable for QTextEdit rich-text mode.

    If maxLines is not -1, this defines a maximum number of lines
    which will be contained in the html output.
    '''
    out = [HTML_HEADER]
    in_stat_data = True
    lines = s.split('\n')
    if maxLines != -1 and len(lines) > maxLines:
        lines = lines[:maxLines]
        lines.extend( [ '[...]', MSG_BIG_DIFF ] )

    for l in lines:
        lesc = html.escape(l)
        if in_stat_data and len(l) and l[0] == ' ':
            if re_stat_line.match(l):
                mo = cast(Match[str], re_stat_line.match(l))
                lout = mo.group(1)
                if len(mo.group(2)):
                    lout += TAG_GREEN_OPEN + mo.group(2) + TAG_GREEN_CLOSE
                if len(mo.group(3)):
                    lout += TAG_RED_OPEN + mo.group(3) + TAG_RED_CLOSE
                out.append(mo.group(4) + lout + TAG_NL)
            else:
                out.append(l + TAG_NL)
            continue
        else:
            in_stat_data = False

        if re_highlight.match(l):
            out.append(TAG_YELLOW_OPEN + l + TAG_YELLOW_CLOSE + TAG_NL)
            continue

        if re_hunk_header.match(l):
            mo = cast(Match[str], re_hunk_header.match(l))
            out.append(TAG_BLUE_OPEN + mo.group(1) + TAG_BLUE_CLOSE + mo.group(2) + TAG_NL)
            continue

        if re_line_added.match(l):
            out.append(TAG_GREEN_OPEN + lesc + TAG_GREEN_CLOSE + TAG_NL)
            continue

        if re_line_removed.match(l):
            out.append(TAG_RED_OPEN + lesc + TAG_RED_CLOSE + TAG_NL)
            continue

        out.append(lesc + TAG_NL)

    out.append(HTML_FOOTER)

    return '\n'.join(out)


def handle_cr_in_text(s: str) -> str:
    '''Handle CR in output text: the line after the CR overwrite the previous line'''
    lines = s.split('\n')
    out_lines = []
    for l in lines:
        if '\r' in l:
            if l[-1] == '\r':
                l = l[:-1]

            sublines = l.split('\r')
            l = sublines[-1]
        out_lines.append(l)

    return '\n'.join(out_lines)


def anonymise_git_url(url: str) -> str:
    # clear username from url
    if not url:
        return url

    if '://' in url:
        if url.startswith('ssh') or url.startswith('git'):
            # for ssh, neutralize username
            url = set_username_on_git_url('username', url or '')
        else:
            # for http, file: and direct clones, strip username
            url = set_username_on_git_url('', url or '')
    else:
        # scp-like ssh url syntax, see man git-clone, section url
        # quote: "the url is recognized if there is no / before the fist colon"
        # example: user@host.xz:path/to/repo.git/
        if ':' in url and (url.find('/') == -1 or url.find('/') > url.find(':')):
            url = set_username_on_git_url('username', url)

    return url


reWinPath = re.compile('^[a-zA-Z]:\\.*')

def set_username_on_git_url(username: str, url: str) -> str:
    '''Take a git url and add or substitute the username part. Returns the new url.

    When url refers to the filesystem, nothing is done on it

    If username is empty, strips the username from the url.
    '''
    if url.lower().startswith('file://'):
        return url

    if '://' in url:
        server_start = url.index('://')+3

    else:
        if url.startswith('/'):
            # linux absolute path
            return url

        if reWinPath.match(url):
            # windows absolute path
            return url

        if not ':' in url:
            # this can not be a scp like syntax, this is a path
            return url

        # scp-like ssh url syntax, see man git-clone, section url
        # quote: "the url is recognized if there is no / before the fist colon"
        # example: user@host.xz:path/to/repo.git/
        if url.find('/') < url.find(':'):
            # this is not a url, this is still a path containing a ':'
            return url

        # ok, this is a scp-like url
        server_start = 0


    try:
        server_end   = url.index('/', server_start)
    except ValueError:
        # unlikely but possible
        server_end = len(url)

    server = url[server_start:server_end]
    if '@' in server:
        server = server.split('@', 1)[1]

    if len(username):
        server = username + '@' + server

    return url[:server_start] + server + url[server_end:]


def deleteDirList(dirList: Iterable[str], ignoreDirDoesNotExist: bool = True) -> str:
    '''Try hard to delete directories listed in input.

    Returns:
        * empty string when everything goes fine.
        * error description when something fails
    '''
    errMsg = ''


    for folder in dirList:
        if not os.path.exists(folder) and ignoreDirDoesNotExist:
            continue

        result = deleteDir(folder)
        if result:
            errMsgHeader = 'Error while deleting: %s\n' % folder
            newErrMsg = errMsgHeader + result + '\n'
            warn(newErrMsg)
            errMsg += newErrMsg

    return errMsg


def deleteDir(folder: str) -> str:
    '''Try hard to delete the input directory.

    When files in the directory are read-only, the initial deletion will fail but this
    function will retry after setting the files as writeable.

    Returns:
        * empty string when everything goes fine.
        * error description when something fails
    '''
    pth = pathlib.Path(folder)

    # check if .git directory is present and unhide it to help deletion
    try:
        _dirContent = list(pth.iterdir())
    except PermissionError:
        # nothing we can do with this directory
        return 'Permission error\n'

    local_error = []

    def rmtree_onerror(_function: Any, path: Any, _excinfo: Any) -> None:
        try:
            # deletion errors are usually because a file is readonly
            os.chmod(path, stat.S_IWRITE)
            os.unlink(path)
        except Exception as exc:
            # exception again !
            local_error.append('''On path %s
    %s''' % (str(path), str(exc)))
            warn(local_error[-1])

    dbg('deleteDirList() - Calling rmtree on %s' % str(pth))
    shutil.rmtree(pth, onerror=rmtree_onerror)

    if len(local_error) > 2:
        local_error = local_error[:2] + ['And more errors...\n']

    return '\n'.join(local_error)


def add_suffix_if_missing(fname: str, ext: str) -> str:
    '''Add the suffix if not already present in the string, and return it'''
    if fname.endswith(ext):
        return fname

    return fname + ext


def extractInt(s: str, intAtBeginning: bool = True) -> int:
    '''Parse the string s which begins by an integer and is followed by some other characters
    extract the integer part and convert it to integer

    If string is empty or contains no integer, just return 0

    This allows natural order sorting.
    '''
    idxStart = 0
    if not intAtBeginning:
        while idxStart < len(s) and not s[idxStart].isdigit():
            idxStart += 1

    idxEnd = idxStart
    while idxEnd < len(s) and s[idxEnd].isdigit():
        idxEnd += 1
    intPart = s[idxStart:idxEnd]
    if len(intPart) == 0:
        return 0
    return int(intPart)


def istrcmp(s1: str, s2: str) -> bool:
    '''Case-insensitive string compare'''
    return s1.lower() < s2.lower()


reGitAuthFailure = re.compile('^' + GIT_AUTH_FAILURE_MARKER, re.MULTILINE)
def hasGitAuthFailureMsg(gitlog: str) -> bool:
    '''Return true if logs contains a git authentication failure message'''
    return bool(reGitAuthFailure.search(gitlog))


GIT_CMD_WITH_NETWORK_AUTH = [ 'fetch', 'pull', 'push', 'ls-remote', 'clone']

def isGitCommandRequiringAuth(cmdline: Sequence[str]) -> bool:
    '''Return True if the git command line uses a command requiring network authentication'''
    if len(cmdline) < 2:
        return False

    git_cmd = cmdline[1]
    if cmdline[1] == '-C' and len(cmdline) >= 3:
        git_cmd = cmdline[3]

    if git_cmd in GIT_CMD_WITH_NETWORK_AUTH:
        return True

    return False



def treeWidgetFlatIterator(item_or_tree: Union[QTreeWidget, QTreeWidgetItem]) -> List[QTreeWidgetItem]:
    '''Iterates through either the QTreeWidget top-level items or through the items of the Item.

    This allows to use the same recursive function on a tree or an item.
    '''
    if isinstance(item_or_tree, QTreeWidget):
        return [cast(QTreeWidgetItem, item_or_tree.topLevelItem(idx)) for idx in range(item_or_tree.topLevelItemCount())]

    return [item_or_tree.child(idx) for idx in range(item_or_tree.childCount())]


def treeWidgetDeepIterator(treeWidget: QTreeWidget) -> Iterator[QTreeWidgetItem]:
    '''Returns all QTreeWidgetItems in a depth-first-search.
    '''
    def yieldChild(item: QTreeWidgetItem) -> Iterator[QTreeWidgetItem]:
        for itemChildIdx in range(item.childCount()):
            itemChild = item.child(itemChildIdx)
            yield itemChild
            yield from yieldChild(itemChild)

    for topItemIdx in range(treeWidget.topLevelItemCount()):
        topItem = treeWidget.topLevelItem(topItemIdx)
        assert topItem
        yield topItem
        yield from yieldChild(topItem)


#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from typing import Union, Any, List

import unittest

from PySide2.QtWidgets import QApplication, QTreeWidget, QTreeWidgetItem

from src.mg_repo_info import MgRepoInfo
from src.mg_utils import treeWidgetFlatIterator, treeWidgetDeepIterator
from src.mg_dialog_git_switch_delete_branch import analyseRepoBranchOrTagInfo, fillBranchTagInfo, \
    GroupingBy, buildRepoBranchInfo, stripOrigin, branchNameIsPresentInRemote, remoteBranchesForBranchName, applyFilterToTree


class TestGitSwitchBranchUtils(unittest.TestCase):

    def setUp(self):
        if QApplication.instance():
            self.app = QApplication.instance()
            return

        self.app = QApplication([])

    def tearDown(self):
        self.app = None

    def testAnanlyseRepoBranchInfo(self):
        self.assertEqual(analyseRepoBranchOrTagInfo([]), [])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['l_br'], ['r_br'])
        ]), [
            ('l_br', 1, 'local', ['repo1']),
            ('r_br', 1, 'remote', ['repo1']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['l_br'], ['r_br']),
            ('repo2', ['l_br'], ['r_br']),
        ]), [
            ('l_br', 2, 'local', ['repo1', 'repo2']),
            ('r_br', 2, 'remote', ['repo1', 'repo2']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['master'], ['master']),
        ]), [
            ('master', 1, 'local and remote', ['repo1']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['master'], ['master']),
            ('repo2', ['master'], ['master']),
        ]), [
            ('master', 2, 'local and remote', ['repo1', 'repo2']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['l_br', 'master'], ['r_br', 'master']),
        ]), [
            ('l_br', 1, 'local', ['repo1']),
            ('r_br', 1, 'remote', ['repo1']),
            ('master', 1, 'local and remote', ['repo1']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['br1'], []),
            ('repo2', [], ['br1']),
        ]), [
            ('br1', 2, 'local and remote', ['repo1', 'repo2']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', ['br1'], []),
            ('repo2', ['br1'], ['br1']),
        ]), [
            ('br1', 2, 'local and remote', ['repo1', 'repo2']),
        ])

        self.assertEqual( analyseRepoBranchOrTagInfo([
            ('repo1', [], ['br1']),
            ('repo2', ['br1'], ['br1']),
        ]), [
            ('br1', 2, 'local and remote', ['repo1', 'repo2']),
        ])

        self.assertEqual(analyseRepoBranchOrTagInfo([
            ('repo1', [], ['br1']),
            ('repo2', ['br1'], ['br1']),
        ]), [
            ('br1', 2, 'local and remote', ['repo1', 'repo2']),
        ])


    def test_crash_during_switch_branch(self):

        class StubRepo: pass
        repo = StubRepo()
        repo.name = 'crash_during_switch_branch'
        repo.branches_local = [
            'dev',
            'feat/Grace_v4_demo',
            'feat/first_client_server_demo',
            'master',
        ]
        repo.branches_remote = [
            'origin/dev',
            'origin/feat/Grace_v4_demo',
            'origin/feat/new_GraceClient_exe_with_pyinstaller',
            'origin/int',
            'origin/master',
            'origin/rel/v0.5.0',
            'origin_FRSA_IP2/HEAD -> origin_FRSA_IP2/master',
            'origin_FRSA_IP2/feat/first_client_server_demo',
            'origin_FRSA_IP2/master',
            'origin_WW2/master',
        ]

        # collect branch names
        repoBranchInfo = buildRepoBranchInfo([repo])
        repoItemInfo = analyseRepoBranchOrTagInfo(repoBranchInfo)


    def test_buildRepoBranchInfo(self):
        class StubRepo: pass
        repo = StubRepo()
        repo.name = 'repo1'
        repo.branches_local = [
            'dev',
            'master',
        ]
        repo.branches_remote = [
            'origin/dev',
            'origin/master',
        ]

        # strips origin by default
        self.assertEqual(buildRepoBranchInfo([repo]),
                         [('repo1', ['dev', 'master'], ['dev', 'master'])]
                         )

        repo.branches_remote = [
            'origin/dev',
            'origin2/master',
        ]
        self.assertEqual(buildRepoBranchInfo([repo]),
                         [('repo1', ['dev', 'master'], ['dev', 'master'])]
                         )

        repo.branches_remote = [
            'origin/dev',
            'origin/master',
            'origin2/master',
        ]
        self.assertEqual(buildRepoBranchInfo([repo]),
                         [('repo1', ['dev', 'master'], ['dev', 'master'])]
                         )


    def test_stripOrigin(self):
        self.assertEqual(stripOrigin( [
            'origin/dev',
            'origin/master',
        ] ), [
            'dev',
            'master'
        ] )

        self.assertEqual(stripOrigin( [
            'origin/dev',
            'origin2/master',
        ] ), [
            'dev',
            'master'
        ] )

        self.assertEqual(stripOrigin( [
                'origin/dev',
                'origin/master',
                'origin2/master',
        ] ), [
            'dev',
            'master',
        ] )


    def test_BranchNameIsPresentInRemote(self):
        self.assertEqual(branchNameIsPresentInRemote('toto', ['origin/toto']), True)
        self.assertEqual(branchNameIsPresentInRemote('toto', ['origin/titi']), False)


    def test_remoteBranchesForBranchName(self):
        self.assertEqual(remoteBranchesForBranchName('toto', ['origin/toto']), ['origin/toto'])
        self.assertEqual(remoteBranchesForBranchName('titi', ['origin/toto']), [])
        self.assertEqual(remoteBranchesForBranchName('toto', ['origin/toto', 'origin2/toto']), ['origin/toto', 'origin2/toto'])


    def extractTreeStructure(self, treeWidget: QTreeWidget, visibleOnly: bool = False) -> Any:
        '''Return the content of the tree widget as a tree structure'''
        treeStructure = []

        def addChildren(item_or_tree: Union[QTreeWidgetItem, QTreeWidget], myTreeStruct: List[Any]):
            for childItem in treeWidgetFlatIterator(item_or_tree):
                if visibleOnly and childItem.isHidden():
                    continue
                myTreeStruct.append( (childItem.text(0).strip() or childItem.text(1).strip(), []) )
                addChildren( childItem, myTreeStruct[-1][1] )

        addChildren(treeWidget, treeStructure)

        return treeStructure

    def testFillBranchInfo(self):
        tree = QTreeWidget()
        fillBranchTagInfo([], tree, GroupingBy.NONE)
        self.assertEqual(tree.topLevelItemCount(), 0)
        tree.clear()

        branchList = [
            ('l_br', 1, 'local', ['repo1']),
            ('master', 1, 'local and remote', ['repo1']),
            ('r_br', 1, 'remote', ['repo1']),
        ]
        fillBranchTagInfo(branchList, tree, GroupingBy.NONE)
        self.assertEqual( self.extractTreeStructure(tree), [
            ('l_br', [('repo1', [])]),
            ('master', [('repo1', [])]),
            ('r_br', [('repo1', [])]),
        ])


        branchList = [
            ('feat/l_br1', 1, 'local', ['repo1']),
            ('feat/l_br2', 1, 'local', ['repo1']),
            ('feat/l_br2/titi', 1, 'local', ['repo1']),
            ('master', 1, 'local and remote', ['repo1']),
            ('xxx/r_br', 1, 'remote', ['repo1']),
        ]
        fillBranchTagInfo(branchList, tree, GroupingBy.NONE)
        self.assertEqual(self.extractTreeStructure(tree), [
            ('feat/l_br1', [('repo1', [])]),
            ('feat/l_br2', [('repo1', [])]),
            ('feat/l_br2/titi', [('repo1', [])]),
            ('master', [('repo1', [])]),
            ('xxx/r_br', [('repo1', [])]),
        ])

        branchList = [
            ('feat/l_br1', 1, 'local', ['repo1']),
            ('feat/l_br2', 1, 'local', ['repo1']),
            ('feat/l_br2/titi', 1, 'local', ['repo1']),
            ('master', 1, 'local and remote', ['repo1']),
            ('xxx/r_br', 1, 'remote', ['repo1']),
        ]
        fillBranchTagInfo(branchList, tree, GroupingBy.NAME)
        self.assertEqual(self.extractTreeStructure(tree), [
            ('feat', [
                ('l_br1', [('repo1', [])]),
                ('l_br2', [
                    ('repo1', []),
                    ('titi', [('repo1', [])]),
                ]),
             ]),
            ('master', [('repo1', [])]),
            ('xxx', [
                ('r_br', [('repo1', [])]),
            ]),
        ])


    def testItemIterator(self):
        tree = QTreeWidget()
        self.assertEqual(list(treeWidgetDeepIterator(tree)), [])


        branchList = [
            ('feat/l_br1', 1, 'local', ['repo1']),
            ('feat/l_br2', 1, 'local', ['repo1']),
            ('feat/l_br2/titi', 1, 'local', ['repo1']),
            ('master', 1, 'local and remote', ['repo1']),
            ('xxx/r_br', 1, 'remote', ['repo1']),
        ]
        fillBranchTagInfo(branchList, tree, GroupingBy.NAME)
        self.assertEqual(self.extractTreeStructure(tree), [
            ('feat', [
                ('l_br1', [
                    ('repo1', [])
                ]),
                ('l_br2', [
                    ('repo1', []),
                    ('titi', [
                        ('repo1', [])
                    ]),
                ]),
            ]),
            ('master', [('repo1', [])]),
            ('xxx', [
                ('r_br', [('repo1', [])]),
            ]),
        ])
        self.assertEqual(list(item.text(0) for item in treeWidgetDeepIterator(tree)), [
            'feat', 'l_br1', '', 'l_br2', '', 'titi', '', 'master', '', 'xxx', 'r_br', ''
        ])


    def testApplyFilterToTree(self):
        tree = QTreeWidget()
        branchList = [
            ('feat/l_br1', 1, 'local', ['repo1']),
            ('feat/l_br2', 1, 'local', ['repo1']),
            ('feat/l_br2/titi', 1, 'local', ['repo1']),
            ('master', 1, 'local and remote', ['repo1']),
            ('xxx/r_br', 1, 'remote', ['repo1']),
        ]
        fillBranchTagInfo(branchList, tree, GroupingBy.NAME)
        normalTreeStructure = [
                ('feat', [
                ('l_br1', [
                    ('repo1', [])
                ]),
                ('l_br2', [
                    ('repo1', []),
                    ('titi', [
                        ('repo1', [])
                    ]),
                ]),
            ]),
            ('master', [
                ('repo1', [])
            ]),
            ('xxx', [
                ('r_br', [
                    ('repo1', [])
                ]),
            ]),
        ]
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), normalTreeStructure)

        # hide everything
        applyFilterToTree(tree, 'yyyyyy')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), [])

        # restore everything
        applyFilterToTree(tree, '')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), normalTreeStructure)

        # more filtering
        applyFilterToTree(tree, 'br1')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), [
            ('feat', [
                ('l_br1', [
                    ('repo1', [])
                ]),
            ])
        ])


        applyFilterToTree(tree, 'feat')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), [
            ('feat', [
                ('l_br1', [
                    ('repo1', [])
                ]),
                ('l_br2', [
                    ('repo1', []),
                    ('titi', [('repo1', [])]),
                ]),
            ]),
        ])


        applyFilterToTree(tree, 'l_br')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), [
            ('feat', [
                ('l_br1', [
                    ('repo1', [])
                ]),
                ('l_br2', [
                    ('repo1', []),
                    ('titi', [('repo1', [])]),
                ]),
            ]),
        ])

        applyFilterToTree(tree, 'titi')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), [
            ('feat', [
                ('l_br2', [
                    ('titi', [('repo1', [])]),
                ]),
            ]),
        ])

        applyFilterToTree(tree, 'master')
        self.assertEqual(self.extractTreeStructure(tree, visibleOnly=True), [
            ('master', [
                ('repo1', [])
            ]),
        ])



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_git_commit.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree
from src.mg_button_history import MgButtonHistory


class Ui_GitCommit(object):
    def setupUi(self, GitCommit: QDialog) -> None:
        if not GitCommit.objectName():
            GitCommit.setObjectName(u"GitCommit")
        GitCommit.resize(863, 423)
        self.verticalLayout_2 = QVBoxLayout(GitCommit)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.groupBox = QGroupBox(GitCommit)
        self.groupBox.setObjectName(u"groupBox")
        self.gridLayout = QGridLayout(self.groupBox)
        self.gridLayout.setObjectName(u"gridLayout")
        self.textEditCommitMessage = QTextEdit(self.groupBox)
        self.textEditCommitMessage.setObjectName(u"textEditCommitMessage")

        self.gridLayout.addWidget(self.textEditCommitMessage, 0, 0, 1, 1)

        self.buttonHistoryCommit = MgButtonHistory(self.groupBox)
        self.buttonHistoryCommit.setObjectName(u"buttonHistoryCommit")

        self.gridLayout.addWidget(self.buttonHistoryCommit, 0, 1, 1, 1)

        self.checkBoxPushToRemote = QCheckBox(self.groupBox)
        self.checkBoxPushToRemote.setObjectName(u"checkBoxPushToRemote")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.checkBoxPushToRemote.sizePolicy().hasHeightForWidth())
        self.checkBoxPushToRemote.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.checkBoxPushToRemote, 1, 0, 1, 1)


        self.verticalLayout_2.addWidget(self.groupBox)

        self.groupBox_2 = QGroupBox(GitCommit)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout = QVBoxLayout(self.groupBox_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.labelRepoSelected = QLabel(self.groupBox_2)
        self.labelRepoSelected.setObjectName(u"labelRepoSelected")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.labelRepoSelected.sizePolicy().hasHeightForWidth())
        self.labelRepoSelected.setSizePolicy(sizePolicy1)
        self.labelRepoSelected.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.labelRepoSelected)

        self.pushButtonAdjustRepoList = QPushButton(self.groupBox_2)
        self.pushButtonAdjustRepoList.setObjectName(u"pushButtonAdjustRepoList")
        sizePolicy2 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(1)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.pushButtonAdjustRepoList.sizePolicy().hasHeightForWidth())
        self.pushButtonAdjustRepoList.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.pushButtonAdjustRepoList)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.treeWidgetRepoList = MgRepoTree(self.groupBox_2)
        self.treeWidgetRepoList.setObjectName(u"treeWidgetRepoList")
        self.treeWidgetRepoList.setAlternatingRowColors(True)
        self.treeWidgetRepoList.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setSortingEnabled(True)

        self.verticalLayout.addWidget(self.treeWidgetRepoList)


        self.verticalLayout_2.addWidget(self.groupBox_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.buttonBox = QDialogButtonBox(GitCommit)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.horizontalLayout_4.addWidget(self.buttonBox)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        QWidget.setTabOrder(self.textEditCommitMessage, self.buttonHistoryCommit)
        QWidget.setTabOrder(self.buttonHistoryCommit, self.checkBoxPushToRemote)
        QWidget.setTabOrder(self.checkBoxPushToRemote, self.pushButtonAdjustRepoList)
        QWidget.setTabOrder(self.pushButtonAdjustRepoList, self.treeWidgetRepoList)

        self.retranslateUi(GitCommit)
        self.buttonBox.accepted.connect(GitCommit.accept)
        self.buttonBox.rejected.connect(GitCommit.reject)

        QMetaObject.connectSlotsByName(GitCommit)
    # setupUi

    def retranslateUi(self, GitCommit: QDialog) -> None:
        GitCommit.setWindowTitle(QCoreApplication.translate("GitCommit", u"Git Commit", None))
        self.groupBox.setTitle(QCoreApplication.translate("GitCommit", u"Message", None))
        self.buttonHistoryCommit.setText("")
        self.checkBoxPushToRemote.setText(QCoreApplication.translate("GitCommit", u"Push to remote", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("GitCommit", u"Targeted repositories", None))
        self.labelRepoSelected.setText(QCoreApplication.translate("GitCommit", u"3 repositories selected", None))
        self.pushButtonAdjustRepoList.setText(QCoreApplication.translate("GitCommit", u"Adjust Repository list", None))
        ___qtreewidgetitem = self.treeWidgetRepoList.headerItem()
        ___qtreewidgetitem.setText(4, QCoreApplication.translate("GitCommit", u"Last Remote Syncho", None));
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("GitCommit", u"Status", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("GitCommit", u"Head", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("GitCommit", u"Git Repo Path", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("GitCommit", u"hidden", None));
    # retranslateUi




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


from typing import List
import logging

from PySide2.QtWidgets import QDialog, QWidget, QFileDialog, QMessageBox, QApplication
from PySide2.QtCore import Qt

import src.mg_config as mgc
from src.mg_repo_info import MgRepoInfo, MultiRepo
from src.gui.ui_export_csv import Ui_dialogCsvExport

logger = logging.getLogger('mg_dialog_export_mgit')
dbg = logger.debug


def runDialogExportCsv(window: QWidget, multiRepo: MultiRepo) -> None:
    """
    Dialog to export current repository state to a multigit file.
    """
    dbg('runDialogExportCsv')
    allRepos = multiRepo.repo_list
    if len(allRepos) == 0:
        QMessageBox.warning(window, 'No repository defined',
                            'Warning: can not export when base directory is not defined.')
        return

    # trigger an asynchronous all repo refresh to shorten the global time when retrieving the CSV information
    for repo_info in allRepos:
        repo_info.ensure_head_and_url_and_commit_date(blocking=False)

    dlg = QDialog(window)
    dlg.setWindowFlags(dlg.windowFlags() & ~Qt.WindowContextHelpButtonHint)
    dlg.setWindowTitle('Choose CSV fields')
    ui = Ui_dialogCsvExport()
    ui.setupUi(dlg)

    dbg('slotExportCsv() - Showing dialog for CSV fields')
    result = dlg.exec()
    dbg('slotExportCsv() - dialog done')

    if not result:
        return

    lastdir = mgc.get_config_instance().lruGetFirst(mgc.CONFIG_LAST_OPENED) or ''
    csvFname, _filter = QFileDialog.getSaveFileName(window, 'Choose CSV file to export', lastdir, '*.csv')
    if not csvFname:
        return

    dbg('slotExportCsv() - processing events')
    QApplication.processEvents()  # give a chance to asynchronous jobs to complete

    fieldsRequested = {
        'head':        ui.checkHead.isChecked(),
        'url':         ui.checkUrl.isChecked(),
        'path':        ui.checkPath.isChecked(),
        'commit_sha1': ui.checkCommitSha1.isChecked(),
        'commit_date': ui.checkCommitDate.isChecked(),
        'branch':      ui.checkCurrentBranch.isChecked(),
        'tag':         ui.checkCurrentTag.isChecked(),
    }
    multiRepo.exportCsv(csvFname, fieldsRequested)
    QMessageBox.information(window, "Export configuration file", "Data exported successfully to {}".format(csvFname))




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import Optional, List, Sequence

import logging, enum

from PySide2.QtCore import QObject, Qt
from PySide2.QtWidgets import QMessageBox

from src.mg_utils import isGitCommandRequiringAuth

logger = logging.getLogger('mg_auth_failure_mgr')
dbg = logger.debug
warn = logger.warning

AUTH_FAILED_LIMIT_BEFORE_WARNING_USER = 2

class ActionAfterManyAuthFailure(enum.Enum):
    ReportWithDialog = enum.auto()
    IgnoreAllAuthErrors = enum.auto()
    BlockFutureGitcommand = enum.auto()
    DialogQuestionInProgress = enum.auto()


class MgAuthFailureMgr(QObject):

    __instance: Optional['MgAuthFailureMgr'] = None
    nbGitAuthFail: int
    actionAfterManyAuthFailure: ActionAfterManyAuthFailure

    def __init__(self) -> None:
        super().__init__()
        dbg('instance created')
        MgAuthFailureMgr.__instance = self
        MgAuthFailureMgr.newSession()


    @staticmethod
    def shouldStopBecauseAuthFailureInProgress(cmdline: Sequence[str]) -> bool:
        if not isGitCommandRequiringAuth(cmdline):
            ret = False
        else:
            ret = (MgAuthFailureMgr.instance().actionAfterManyAuthFailure == ActionAfterManyAuthFailure.BlockFutureGitcommand)

        dbg(f'shouldStopBecauseAuthFailureInProgress() -> {ret}')
        return ret


    @staticmethod
    def gitAuthFailed(cmdline: str) -> None:
        '''Called each time a git authentication fails for a git command'''
        self = MgAuthFailureMgr.instance()
        self.nbGitAuthFail += 1
        warn(f'Git authentication failure: {self.nbGitAuthFail} on {AUTH_FAILED_LIMIT_BEFORE_WARNING_USER} authorized')
        if self.nbGitAuthFail > AUTH_FAILED_LIMIT_BEFORE_WARNING_USER \
                and self.actionAfterManyAuthFailure == ActionAfterManyAuthFailure.ReportWithDialog:
            self.actionAfterManyAuthFailure = ActionAfterManyAuthFailure.DialogQuestionInProgress
            msg = f'We have already detected {self.nbGitAuthFail} authentication failures.'
            msg += ' Be careful that too many authentication failures may lock your account.'
            msg += '<p>To update your git login/password, open the <i>Windows Credentials Manager (in french: Gérer les mots de passe réseau</i>)'
            msg += '<p>What to you want to do next ?'

            msgBox = QMessageBox(QMessageBox.Icon.Warning, "Many authentication failures", '')
            msgBox.setTextFormat(Qt.TextFormat.RichText)
            msgBox.setText(msg)
            buttonContinue = msgBox.addButton('Continue', QMessageBox.ButtonRole.AcceptRole)
            buttonContinueAndIgnore = msgBox.addButton('Continue && Ignore all authentication failures', QMessageBox.ButtonRole.AcceptRole)
            buttonCancelAll = msgBox.addButton('Cancel all git operations', QMessageBox.ButtonRole.AcceptRole)
            msgBox.exec()
            buttonSelected = msgBox.clickedButton()
            if buttonSelected == buttonCancelAll:
                self.actionAfterManyAuthFailure = ActionAfterManyAuthFailure.BlockFutureGitcommand
            elif buttonSelected == buttonContinueAndIgnore:
                self.actionAfterManyAuthFailure = ActionAfterManyAuthFailure.IgnoreAllAuthErrors
            elif buttonSelected == buttonContinue:
                self.actionAfterManyAuthFailure = ActionAfterManyAuthFailure.ReportWithDialog


    @classmethod
    def instance(cls) -> 'MgAuthFailureMgr':
        '''Return an existing instance of this class, creating one if necessary'''
        if cls.__instance is None:
            cls.__instance = MgAuthFailureMgr()

        return cls.__instance

    @classmethod
    def newSession(cls) -> None:
        '''A new session of git commands is started, so reset all counters and actions'''
        self = cls.instance()
        self.nbGitAuthFail = 0
        self.actionAfterManyAuthFailure = ActionAfterManyAuthFailure.ReportWithDialog





# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_git_run_command.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree
from src.mg_button_history import MgButtonHistory

# import multigit_resources_rc

class Ui_GitRunCommand(object):
    def setupUi(self, GitRunCommand: QDialog) -> None:
        if not GitRunCommand.objectName():
            GitRunCommand.setObjectName(u"GitRunCommand")
        GitRunCommand.resize(869, 384)
        self.verticalLayout_2 = QVBoxLayout(GitRunCommand)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.lineEditGitText = QLineEdit(GitRunCommand)
        self.lineEditGitText.setObjectName(u"lineEditGitText")
        self.lineEditGitText.setEnabled(True)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEditGitText.sizePolicy().hasHeightForWidth())
        self.lineEditGitText.setSizePolicy(sizePolicy)
        font = QFont()
        font.setFamily(u"Consolas")
        self.lineEditGitText.setFont(font)
        self.lineEditGitText.setAlignment(Qt.AlignCenter)
        self.lineEditGitText.setReadOnly(True)

        self.gridLayout.addWidget(self.lineEditGitText, 0, 1, 1, 1)

        self.lineEditGitCmd = QLineEdit(GitRunCommand)
        self.lineEditGitCmd.setObjectName(u"lineEditGitCmd")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(11)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.lineEditGitCmd.sizePolicy().hasHeightForWidth())
        self.lineEditGitCmd.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.lineEditGitCmd, 0, 2, 1, 1)

        self.label = QLabel(GitRunCommand)
        self.label.setObjectName(u"label")
        self.label.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.historyButton = MgButtonHistory(GitRunCommand)
        self.historyButton.setObjectName(u"historyButton")
        icon = QIcon()
        icon.addFile(u":/img/icons8-history-64.png", QSize(), QIcon.Normal, QIcon.Off)
        self.historyButton.setIcon(icon)

        self.gridLayout.addWidget(self.historyButton, 0, 3, 1, 1)


        self.verticalLayout_2.addLayout(self.gridLayout)

        self.widget = QWidget(GitRunCommand)
        self.widget.setObjectName(u"widget")
        self.widget.setMinimumSize(QSize(0, 30))

        self.verticalLayout_2.addWidget(self.widget)

        self.groupBox_2 = QGroupBox(GitRunCommand)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout = QVBoxLayout(self.groupBox_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.labelRepoSelected = QLabel(self.groupBox_2)
        self.labelRepoSelected.setObjectName(u"labelRepoSelected")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy2.setHorizontalStretch(1)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.labelRepoSelected.sizePolicy().hasHeightForWidth())
        self.labelRepoSelected.setSizePolicy(sizePolicy2)
        self.labelRepoSelected.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.labelRepoSelected)

        self.pushButtonAdjustRepoList = QPushButton(self.groupBox_2)
        self.pushButtonAdjustRepoList.setObjectName(u"pushButtonAdjustRepoList")
        sizePolicy3 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.pushButtonAdjustRepoList.sizePolicy().hasHeightForWidth())
        self.pushButtonAdjustRepoList.setSizePolicy(sizePolicy3)

        self.horizontalLayout.addWidget(self.pushButtonAdjustRepoList)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.treeWidgetRepoList = MgRepoTree(self.groupBox_2)
        self.treeWidgetRepoList.setObjectName(u"treeWidgetRepoList")
        self.treeWidgetRepoList.setAlternatingRowColors(True)
        self.treeWidgetRepoList.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setSortingEnabled(True)

        self.verticalLayout.addWidget(self.treeWidgetRepoList)


        self.verticalLayout_2.addWidget(self.groupBox_2)

        self.label_3 = QLabel(GitRunCommand)
        self.label_3.setObjectName(u"label_3")

        self.verticalLayout_2.addWidget(self.label_3)

        self.buttonBox = QDialogButtonBox(GitRunCommand)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout_2.addWidget(self.buttonBox)

        QWidget.setTabOrder(self.lineEditGitText, self.lineEditGitCmd)
        QWidget.setTabOrder(self.lineEditGitCmd, self.historyButton)
        QWidget.setTabOrder(self.historyButton, self.pushButtonAdjustRepoList)
        QWidget.setTabOrder(self.pushButtonAdjustRepoList, self.treeWidgetRepoList)

        self.retranslateUi(GitRunCommand)
        self.buttonBox.accepted.connect(GitRunCommand.accept)
        self.buttonBox.rejected.connect(GitRunCommand.reject)

        QMetaObject.connectSlotsByName(GitRunCommand)
    # setupUi

    def retranslateUi(self, GitRunCommand: QDialog) -> None:
        GitRunCommand.setWindowTitle(QCoreApplication.translate("GitRunCommand", u"Run git command", None))
        self.lineEditGitText.setText(QCoreApplication.translate("GitRunCommand", u"git", None))
        self.label.setText(QCoreApplication.translate("GitRunCommand", u"Git command to run :", None))
#if QT_CONFIG(tooltip)
        self.historyButton.setToolTip(QCoreApplication.translate("GitRunCommand", u"Choose recent multigit file", None))
#endif // QT_CONFIG(tooltip)
        self.historyButton.setText("")
        self.groupBox_2.setTitle(QCoreApplication.translate("GitRunCommand", u"Targeted repositories", None))
        self.labelRepoSelected.setText(QCoreApplication.translate("GitRunCommand", u"3 repositories selected", None))
        self.pushButtonAdjustRepoList.setText(QCoreApplication.translate("GitRunCommand", u"Adjust Repository list", None))
        ___qtreewidgetitem = self.treeWidgetRepoList.headerItem()
        ___qtreewidgetitem.setText(4, QCoreApplication.translate("GitRunCommand", u"Last Remote Syncho", None));
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("GitRunCommand", u"Status", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("GitRunCommand", u"Head", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("GitRunCommand", u"Git Repo Path", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("GitRunCommand", u"hidden", None));
        self.label_3.setText("")
    # retranslateUi




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, TypeVar, Sequence, Optional, Any, Dict

import os, ast, sys
from pathlib import Path
from pprint import pformat
import logging, traceback
import datetime, shutil

from PySide2.QtCore import Qt
from PySide2.QtWidgets import QMessageBox, QApplication
from PySide2.QtGui import QColor

# keys for the configuration file
CONFIG_MAINWINDOW_GEOMETRY = 'MAINWINDOW_GEOMETRY'
CONFIG_LAST_OPENED = 'LAST_OPENED'
CONFIG_TABS_OPENED = 'CONFIG_TABS_OPENED'
CONFIG_TABS_CURRENT = 'CONFIG_TABS_CURRENT'
CONFIG_SPLITTER_STATE = 'SPLITTER_STATE'
CONFIG_SPLITTER_STATE_V2 = 'SPLITTER_STATE_V2'
CONFIG_GIT_AUTODETECT = 'CONFIG_GIT_AUTODETECT'
CONFIG_GIT_MANUAL_PATH = 'CONFIG_GIT_MANUAL_PATH'
CONFIG_TORTOISEGIT_AUTODETECT = 'CONFIG_TORTOISEGIT_AUTODETECT'
CONFIG_TORTOISEGIT_MANUAL_PATH = 'CONFIG_TORTOISEGIT_MANUAL_PATH'
CONFIG_TORTOISEGIT_ACTIVATED = 'CONFIG_TORTOISEGIT_ACTIVATED'
CONFIG_SOURCETREE_AUTODETECT = 'CONFIG_SOURCETREE_AUTODETECT'
CONFIG_SOURCETREE_MANUAL_PATH = 'CONFIG_SOURCETREE_MANUAL_PATH'
CONFIG_SOURCETREE_ACTIVATED = 'CONFIG_SOURCETREE_ACTIVATED'
CONFIG_SUBLIMEMERGE_AUTODETECT = 'CONFIG_SUBLIMEMERGE_AUTODETECT'
CONFIG_SUBLIMEMERGE_MANUAL_PATH = 'CONFIG_SUBLIMEMERGE_MANUAL_PATH'
CONFIG_SUBLIMEMERGE_ACTIVATED = 'CONFIG_SUBLIMEMERGE_ACTIVATED'
CONFIG_GITGUI_AUTODETECT = 'CONFIG_GITGUI_AUTODETECT'
CONFIG_GITGUI_MANUAL_PATH = 'CONFIG_GITGUI_MANUAL_PATH'
CONFIG_GITGUI_ACTIVATED = 'CONFIG_GITGUI_ACTIVATED'
CONFIG_GITK_AUTODETECT = 'CONFIG_GITK_AUTODETECT'
CONFIG_GITK_MANUAL_PATH = 'CONFIG_GITK_MANUAL_PATH'
CONFIG_GITK_ACTIVATED = 'CONFIG_GITK_ACTIVATED'
CONFIG_GITBASH_AUTODETECT = 'CONFIG_GITBASH_AUTODETECT'
CONFIG_GITBASH_MANUAL_PATH = 'CONFIG_GITBASH_MANUAL_PATH'
CONFIG_GITBASH_ACTIVATED = 'CONFIG_GITBASH_ACTIVATED'
CONFIG_EXPLORER_AUTODETECT = 'CONFIG_EXPLORER_AUTODETECT'
CONFIG_EXPLORER_MANUAL_PATH = 'CONFIG_EXPLORER_MANUAL_PATH'
CONFIG_TAG_HISTORY = 'CONFIG_TAG_HISTORY'
CONFIG_GIT_CMD_HISTORY = 'CONFIG_GIT_CMD_HISTORY'
CONFIG_GIT_BRANCH_HISTORY = 'CONFIG_GIT_BRANCH_HISTORY'
CONFIG_DOUBLE_CLICK_ACTION = 'CONFIG_DOUBLE_CLICK_ACTION'
CONFIG_HEAD_COLOR_BRANCH = 'CONFIG_HEAD_COLOR_BRANCH'
CONFIG_HEAD_COLOR_TAG = 'CONFIG_HEAD_COlOR_TAG'
CONFIG_LAST_SHOWN_WHATISNEW = 'CONFIG_LAST_SHOWN_WHATISNEW'
CONFIG_LAST_MGIT_FILE = 'CONFIG_LAST_PROJECT_FILE'
CONFIG_LAST_PROJECT_DIR = 'CONFIG_LAST_PROJECT_DIR'
CONFIG_NB_GIT_PROC = 'CONFIG_NB_GIT_PROC'
CONFIG_LOG_DEBUG = 'CONFIG_LOG_DEBUG'
CONFIG_GIT_COMMIT_HISTORY = 'CONFIG_GIT_COMMIT_HISTORY'
CONFIG_VIEW_TAB_LAST_COMMIT = 'CONFIG_VIEW_TAB_LAST_COMMIT'
CONFIG_VIEW_TAB_MOD_FILES = 'CONFIG_VIEW_TAB_MOD_FILES'
CONFIG_VIEW_COL_SHA1 = 'CONFIG_VIEW_COL_SHA1'
CONFIG_VIEW_COL_URL = 'CONFIG_VIEW_COL_URL'
CONFIG_CLONE_USERNAME = 'CONFIG_CONFIG_CLONE_USERNAME'
CONFIG_FETCH_ON_STARTUP = 'CONFIG_FETCH_ON_STARTUP'
CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN = 'CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN'

DEFAULT_CHECK_UPDATE_FREQUENCY = 30 # check every month

# Format is: 0xAARRGGBB with AA = alpha, RR = red, GG = green, BB = blue
DEFAULT_CONFIG_HEAD_COLOR_BRANCH = QColor(Qt.blue).rgb()
DEFAULT_CONFIG_HEAD_COLOR_TAG = QColor(Qt.darkCyan).rgb()

logger = logging.getLogger('mg_config')
dbg = logger.debug

__CONFIG_INSTANCE = None


def get_config_instance() -> 'MgConfig':
    '''Return an instance of the configuration of Multigit, with the default configuration path'''
    global __CONFIG_INSTANCE

    if sys.platform == 'win32':
        default_config_path = Path(os.environ['USERPROFILE']) / 'AppData/Local/MultiGit/multigit.config'
    else:
        # See https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
        xdg_config_home: str
        if os.environ.get('XDG_CONFIG_HOME', ''):
            xdg_config_home = os.environ['XDG_CONFIG_HOME']
        else:
            xdg_config_home = os.path.expanduser('~/.config')

        default_config_path = Path(xdg_config_home) / 'Multigit/multigit.config'

    if __CONFIG_INSTANCE is None:
        __CONFIG_INSTANCE = MgConfig(str(default_config_path))
        __CONFIG_INSTANCE.load()
    return __CONFIG_INSTANCE


class MgConfig:
    '''MultiGit configuration management'''

    def __init__(self, config_path: str):
        '''Set the configuration filename.
        '''
        self.config_dict: Dict[str, Any] = {}
        self.lru_dict: Dict[str, LRUList] = {}
        self.config_path = Path(config_path)
        self.do_not_save = False

    def load(self) -> None:
        '''Load configuration from the config location'''
        if not self.config_path.exists():
            # start from an empty configuration
            self.config_dict = {}
            return

        with open(self.config_path, 'r', encoding='utf8') as f:
            config_content = f.read()

        try:
            self.config_dict = ast.literal_eval(config_content)
            return
        except Exception as exc:
            self.do_not_save = True

        # we know that configuration loading failed
        logger.error(traceback.format_exc())
        self.config_dict = {}

        msg1 = 'Could not read configuration file'
        msg2 = ('Warning: unable to understand content of configuration file:\n%s\n\n' % self.config_path +
                'Configuration file will not be updated.\n'
                'Please report this to philippe.fremy@idemia.com\n'
                )

        logger.error(msg1)
        logger.error(msg2)

        # if QApplication exists, let's display a warning message
        if QApplication.instance():
            QMessageBox.warning(None, msg1, msg2 )

        invalidFileName = self.config_path.parent / \
                          ('multigit-invalid-config-file-%s.config' % datetime.datetime.now().isoformat().replace(':', '.'))

        msg2 = f'The invalid configuration file has been renamed to:\n{invalidFileName}\nPlease include this file when reporting the problem to philippe.fremy@idemia.com'
        try:
            shutil.move(str(self.config_path), invalidFileName)
            logger.error(msg2)
            if QApplication.instance():
                QMessageBox.warning(None, msg1, msg2)

            # corrupted config file is gone, let's create a new one
            self.do_not_save = False

        except OSError:
            try:
                shutil.copy(self.config_path, invalidFileName)
                logger.error(msg2)
                if QApplication.instance():
                    QMessageBox.warning(None, msg1, msg2)

                # corrupted config file is gone, let's create a new one
                self.do_not_save = False

            except OSError:
                logger.error('Unable to move or copy the faulty file')
                # we could not even copy the file, make sure it is not overwritten
                # by not allowing to save configuration
                pass

                msg2 = (f'To fix the configuration file, you can either open the configuration file and fix the syntax error.'
                       + f'Or you can simply erase the file, MultiGit will create a new empty configuration.\n'
                       + f'\nThe configuration file is :\n{self.config_path}'
                        )
                if QApplication.instance():
                    QMessageBox.warning(None, msg1, msg2)


    def save(self) -> None:
        '''Save configuration into the config location'''
        dbg('Saving config to %s' % self.config_path)
        if not self.config_path.parent.exists():
            self.config_path.parent.mkdir(parents=True)

        if self.do_not_save:
            # we don't want to write a config which we failed to read first, so skip silently
            return

        with open(self.config_path, 'w', encoding='utf8') as f:
            f.write(pformat(self.config_dict, indent=4, width=200))

    def __getitem__(self, config_key: str) -> Any:
        '''Return the content of the configuration key or None if the key does not exist.'''
        return self.config_dict.get(config_key)

    def get(self, config_key: str, default_value: Any = None) -> Any:
        '''Return the content of the configuration key or None if the key does not exist.'''
        return self.config_dict.get(config_key, default_value)

    def __setitem__(self, key: str, value: Any) -> None:
        '''Set a configuration key with a value'''
        if isinstance(value, Path):
            value = str(value)
        self.config_dict[key] = value


    def _lruCreateOrGet(self, config_key: str) -> 'LRUList':
        '''Create a LRU list for a given configuration entry on-demand, or return
        an existing one if it already exists'''
        if not config_key in self.lru_dict:
            value = self.get(config_key) or []
            self.lru_dict[config_key] = LRUList(value)
        return self.lru_dict[config_key]


    def lruGetFirst(self, config_key: str) -> Any:
        '''Return the last lru element of config key config_key'''
        return self._lruCreateOrGet(config_key).getFirst()


    def lruSetRecent(self, config_key: str, value: Any) -> None:
        '''Update the LRU entry for config_key with the given value'''
        lru = self._lruCreateOrGet(config_key)
        lru.setRecent(value)
        self.config_dict[config_key] = lru.asList()


    def lruAsList(self, config_key: str) -> List[Any]:
        '''Return the full content of the LRU entry config_key as a list'''
        return self._lruCreateOrGet(config_key).asList()


class LRUList:
    _l: List[Any]

    def __init__(self, l: Sequence[Any], maxSize: int = 20) -> None:
        self.maxSize = maxSize
        self._l = []
        for v in reversed(l):
            # add them in reverse order, guarantee that no double are presents
            self.setRecent(v)

    def getFirst(self) -> Optional[Any]:
        '''Return the first item of the LRU list, without modifying it.
        Returns None if the list is empty.'''
        if len(self._l):
            return self._l[0]
        else:
            return None

    def setRecent(self, v: Any) -> None:
        # normally, there is zero or one copy
        assert (self._l.count(v) < 2)

        # remove previous instance if any
        if v in self._l:
            self._l.remove(v)

        self._l.insert(0, v)

        # ensure max size
        while len(self._l) > self.maxSize:
            self._l.pop()

    def asList(self) -> List[Any]:
        return self._l[:]



#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


import unittest, os

from src.mg_utils import htmlize_diff, handle_cr_in_text, set_username_on_git_url, add_suffix_if_missing, extractInt, \
    hasGitAuthFailureMsg, isGitCommandRequiringAuth, anonymise_git_url
from src.mg_config import MgConfig
from src.mg_const import MSG_BIG_DIFF
from src.mg_repo_info import match_ahead_behind, is_not_sha1, MgRepoInfo


#######################################################
#           Tests for various utilities
#######################################################


class TestUtilityFunctions(unittest.TestCase):

    def test_anonymise_git_url(self):
        test_data = [
            ("ssh://fremy@some.server.com:29418/some/path/to/repo.git",
             "ssh://username@some.server.com:29418/some/path/to/repo.git"),
            ("ssh://some.server.com:29418/some/path/to/repo.git",
             "ssh://username@some.server.com:29418/some/path/to/repo.git"),
            ("https://fremy@some.server.com:29418/some/path/to/repo.git",
             "https://some.server.com:29418/some/path/to/repo.git"),
            ("https://some.server.com:29418/some/path/to/repo.git",
             "https://some.server.com:29418/some/path/to/repo.git"),
            (r"c:\cloned\from\filesystem",
             r"c:\cloned\from\filesystem",),
            ("file://C:/work/Multigit/Sandbox/",
             "file://C:/work/Multigit/Sandbox/"),

            # ssh equivalent, see man git-clone, sections url
            ('fremy@host.xz:path/to/repo.git/',
             'username@host.xz:path/to/repo.git/'),
        ]

        for before, after in test_data:
            with self.subTest(before) as _:
                self.assertEqual( anonymise_git_url(before), after)


    def test_set_username_to_git_url(self):

        test_data = [
            ("ssh://fremy@some.server.com:29418/some/path/to/repo.git",
                "ssh://totoro@some.server.com:29418/some/path/to/repo.git"),
            ("ssh://some.server.com:29418/some/path/to/repo.git",
                "ssh://totoro@some.server.com:29418/some/path/to/repo.git"),
            ("https://fremy@some.server.com:29418/some/path/to/repo.git",
                "https://totoro@some.server.com:29418/some/path/to/repo.git"),
            ("https://some.server.com:29418/some/path/to/repo.git",
                "https://totoro@some.server.com:29418/some/path/to/repo.git"),
            ("http://fremy@some.server.com:29418/some/path/to/repo.git",
                "http://totoro@some.server.com:29418/some/path/to/repo.git"),
            ("http://some.server.com:29418/some/path/to/repo.git",
                "http://totoro@some.server.com:29418/some/path/to/repo.git"),
            (r"c:\cloned\from\filesystem",
                r"c:\cloned\from\filesystem",),
            (r"cloned\from\filesystem",
             r"cloned\from\filesystem",),
            (r"cloned/from/filesystem",
             r"cloned/from/filesystem",),
            (r"/cloned/from/filesystem",
             r"/cloned/from/filesystem",),
            ("file://C:/work/Multigit/Sandbox/",
                 "file://C:/work/Multigit/Sandbox/"),
        ]

        for before, after in test_data:
            with self.subTest(before) as _:
                self.assertEqual( set_username_on_git_url('totoro', before), after)

        self.assertEqual( set_username_on_git_url('', "ssh://fremy@some.server.com:29418/some/path/to/repo.git"),
                          "ssh://some.server.com:29418/some/path/to/repo.git")
        self.assertEqual( set_username_on_git_url('', "ssh://some.server.com:29418/some/path/to/repo.git"),
             "ssh://some.server.com:29418/some/path/to/repo.git")

    def test_handle_cr_in_text(self):
        self.assertEqual( handle_cr_in_text('abc\n'),               'abc\n')
        self.assertEqual( handle_cr_in_text('abc\ndef\r'),          'abc\ndef')
        self.assertEqual( handle_cr_in_text('abc\ndef\rDEF\r'),     'abc\nDEF')
        self.assertEqual( handle_cr_in_text('abc\ndef\rDEF\r123\n'),'abc\n123\n')


    def test_match_ahead_behind(self):
        self.assertEqual(match_ahead_behind('ahead 33'), (33, 0))
        self.assertEqual(match_ahead_behind('behind 22'), (0, 22))
        self.assertEqual(match_ahead_behind('ahead 33, behind 22'), (33, 22))

    def test_is_not_sha1(self):
        self.assertEqual(is_not_sha1('11'), False)
        self.assertEqual(is_not_sha1('11af'), False)
        self.assertEqual(is_not_sha1('11.af'), True)
        self.assertEqual(is_not_sha1('11.22'), True)
        self.assertEqual(is_not_sha1('11_af'), True)

    def test_re_mod_files(self):
        mo = MgRepoInfo.re_status_mod_files.match('A  toto')
        assert mo
        self.assertEqual(mo.group(2), 'toto')
        mo = MgRepoInfo.re_status_mod_files.match(' D toto')
        assert mo
        self.assertEqual(mo.group(2), 'toto')
        mo = MgRepoInfo.re_status_mod_files.match(' D toto')
        assert mo
        self.assertEqual(mo.group(2), 'toto')
        mo = MgRepoInfo.re_status_mod_files.match('AD "toto titi"')
        assert mo
        self.assertEqual(mo.group(2), '"toto titi"')
        mo = MgRepoInfo.re_status_mod_files.match('AD toto')
        assert mo
        self.assertEqual(mo.group(2), 'toto')
        mo = MgRepoInfo.re_status_mod_files.match(' toto')
        self.assertEqual(mo, None)

        mo = MgRepoInfo.re_status_mod_files.match('A  toto -> titi')
        assert mo
        self.assertEqual(mo.group(2), 'toto -> titi')
        mo = MgRepoInfo.re_status_mod_files.match('AD toto -> titi')
        assert mo
        self.assertEqual(mo.group(2), 'toto -> titi')

    def test_add_suffix_if_missing(self):
        self.assertEqual(add_suffix_if_missing('toto', '.mgit'), 'toto.mgit')
        self.assertEqual(add_suffix_if_missing('toto.mgit', '.mgit'), 'toto.mgit')


    def test_extract_int(self):
        self.assertEqual(extractInt(''), 0)
        self.assertEqual(extractInt('123'), 123)
        self.assertEqual(extractInt('123xxx45'), 123)
        self.assertEqual(extractInt('001xxx45'), 1)
        self.assertEqual(extractInt('xxx45'), 0)
        self.assertEqual(extractInt('xxx45', intAtBeginning=False), 45)
        self.assertEqual(extractInt('001xxx45', intAtBeginning=True), 1)


    def test_hasGitAuthFailureMsg(self):
        gitlog = '''warning: ----------------- SECURITY WARNING ----------------
warning: | TLS certificate verification has been disabled! |
warning: ---------------------------------------------------
warning: HTTPS connections may not be secure. See https://aka.ms/gcmcore-tlsverify for more information.
fatal: Authentication failed for 'https://xxx.yyy.com/toto/Sandbox.git/'
bla bla bla
'''
        self.assertEqual(hasGitAuthFailureMsg(gitlog), True)
        gitlog2 = '''xxfatal: Authentication failed for 'https://xxx.yyy.com/toto/Sandbox.git/' 
    '''
        self.assertEqual(hasGitAuthFailureMsg(gitlog2), False)


    def test_isGitCommandRequiringAuth(self):
        self.assertEqual(isGitCommandRequiringAuth(['git', 'push', '--verbose']), True)
        self.assertEqual(isGitCommandRequiringAuth(['git', 'pull']), True)
        self.assertEqual(isGitCommandRequiringAuth(['git', 'fetch', '--progress']), True)

        for cmdline in [
            ['git.exe', '--version'],
            ['git.exe', '-C', r'C:\work\Multigit\Dev', 'log', '-1'],
            ['git.exe', '-C', r'C:\work\Multigit\Dev', 'remote', '--verbose'],
            ['git.exe', '-C', r'C:\work\Multigit\Dev', 'status', '--porcelain', '--branch'],
            ['git.exe', '-C', r'C:\work\Multigit\Sandbox_http', 'diff', '-u', '--patience', '--stat'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\submodule2', 'checkout', 'master'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'merge', 'dev'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'status', '--porcelain', '--branch'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'tag', '--list', '--sort',
             'creatordate'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo2', 'branch'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo2', 'checkout', 'dev'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo2', 'checkout', 'int'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'tag', 'int_112234', '-F', r'C:\Users\g582619\AppData\Local\Temp\tmp_b_uwll1']
        ]:
            with self.subTest(' '.join(cmdline)):
                self.assertEqual(isGitCommandRequiringAuth(cmdline), False)

        for cmdline in [
            ['git.exe', '-C', r'C:\work\Multigit\Sandbox_http', 'fetch', '--verbose', '--prune'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'fetch', '--prune'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'pull'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'push', '-u', '--progress', 'origin', 'int:int'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo1', 'push', 'origin', 'int_112234'],
            ['git.exe', '-C', r'C:\work\Multigit\tmp\toto2\subdir1\subdir1_repo2', 'fetch', '--prune'],
            ['git.exe', 'clone', '--progress file://C:/work/Multigit/Sandbox/', r'C:\work\Multigit\tmp\submodule2'],
            ['git.exe', 'clone', '--progress file://C:/work/Multigit/Sandbox/', r'C:\work\Multigit\tmp\submodule2'],
        ]:
            with self.subTest(' '.join(cmdline)):
                self.assertEqual(isGitCommandRequiringAuth(cmdline), True)


    def test_mg_config(self):
        try:
            mgc = MgConfig('my_config.cfg')
            mgc['toto'] = { 1:2, 3:4 }
            mgc['titi'] = [ 1, 2, 3, 4]
            mgc['tutu'] = 'coucou'

            assert mgc['toto'] == { 1:2, 3:4 }
            assert mgc['titi'] == [1,2,3,4]
            assert mgc['tutu'] == 'coucou'

            assert mgc['key_not_present'] is  None

            mgc.save()

            mgc = MgConfig('my_config.cfg')
            mgc.load()

            assert mgc['toto'] == { 1:2, 3:4 }
            assert mgc['titi'] == [1,2,3,4]
            assert mgc['tutu'] == 'coucou'

        finally:
            if os.path.exists('my_config.cfg'):
                os.unlink('my_config.cfg')

    def test_htmlize_diff(self):
        sin = ''' test/test_gui.py    | 16 ++++++++++++++++
 test/test_sxtool.py |  6 +-----
 2 files changed, 17 insertions(+), 5 deletions(-)

diff --git a/test/test_gui.py b/test/test_gui.py
index 8125a27..3e51db7 100644
--- a/test/test_gui.py
+++ b/test/test_gui.py
@@ -11,6 +11,22 @@ from PyQt5.QtTest import QTest

 from src.main_form import MainForm
+
+from src.const import CONST_33
+
@@ this is a description @@
this line is plain
- this line is removed
+ this line is added
'''

        sout_ref='''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li { white-space: pre-wrap;font-family:'Courier New'; font-size:8pt; font-weight:600; 
        font-style:normal; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; 
        -qt-block-indent:0; text-indent:0px; }
</style></head>
<body style="font-family:'Courier New'; font-size:8pt; font-weight:400; font-style:normal">
 test/test_gui.py    | 16 <span style="background-color:#ccffcc;">++++++++++++++++</span><br/>
 test/test_sxtool.py |  6 <span style="background-color:#ccffcc;">+</span><span style="background-color:#ffdddd;">-----</span><br/>
 2 files changed, 17 insertions(+), 5 deletions(-)<br/>
<br/>
<span style="background-color:#ffff7f ">diff --git a/test/test_gui.py b/test/test_gui.py</span><br/>
<span style="background-color:#ffff7f ">index 8125a27..3e51db7 100644</span><br/>
<span style="background-color:#ffff7f ">--- a/test/test_gui.py</span><br/>
<span style="background-color:#ffff7f ">+++ b/test/test_gui.py</span><br/>
<span style="background-color:#aaffff;">@@ -11,6 +11,22 @@</span> from PyQt5.QtTest import QTest<br/>
<br/>
 from src.main_form import MainForm<br/>
<span style="background-color:#ccffcc;">+</span><br/>
<span style="background-color:#ccffcc;">+from src.const import CONST_33</span><br/>
<span style="background-color:#ccffcc;">+</span><br/>
<span style="background-color:#aaffff;">@@ this is a description @@</span><br/>
this line is plain<br/>
<span style="background-color:#ffdddd;">- this line is removed</span><br/>
<span style="background-color:#ccffcc;">+ this line is added</span><br/>
<br/>
</body></html>'''

        sout = htmlize_diff(sin)
        if sout != sout_ref:
            for l, lref in zip(sout.split('\n'), sout_ref.split('\n')):
                if l == lref:
                    print('={}\\n'.format(l))
                else:
                    print('-{}\\n'.format(l))
                    print('+{}\\n'.format(lref))
        self.assertEqual(sout, sout_ref)

    def test_htmlize_diff_maxlines(self):
        sin = ''' test/test_gui.py    | 16 ++++++++++++++++
 test/test_sxtool.py |  6 +-----
 2 files changed, 17 insertions(+), 5 deletions(-)

diff --git a/test/test_gui.py b/test/test_gui.py
index 8125a27..3e51db7 100644
--- a/test/test_gui.py
+++ b/test/test_gui.py
@@ -11,6 +11,22 @@ from PyQt5.QtTest import QTest
'''

        sout_ref='''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li { white-space: pre-wrap;font-family:'Courier New'; font-size:8pt; font-weight:600; 
        font-style:normal; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; 
        -qt-block-indent:0; text-indent:0px; }
</style></head>
<body style="font-family:'Courier New'; font-size:8pt; font-weight:400; font-style:normal">
 test/test_gui.py    | 16 <span style="background-color:#ccffcc;">++++++++++++++++</span><br/>
 test/test_sxtool.py |  6 <span style="background-color:#ccffcc;">+</span><span style="background-color:#ffdddd;">-----</span><br/>
 2 files changed, 17 insertions(+), 5 deletions(-)<br/>
<br/>
[...]<br/>
%s<br/>
</body></html>''' % MSG_BIG_DIFF

        sout = htmlize_diff(sin, 4)
        if sout != sout_ref:
            for l, lref in zip(sout.split('\n'), sout_ref.split('\n')):
                if l == lref:
                    print('={}\\n'.format(l))
                else:
                    print('-{}\\n'.format(l))
                    print('+{}\\n'.format(lref))
        self.assertEqual(sout, sout_ref)

        sin = ''

        sout_ref='''<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN" "http://www.w3.org/TR/REC-html40/strict.dtd">
<html><head><meta name="qrichtext" content="1" /><style type="text/css">
p, li { white-space: pre-wrap;font-family:'Courier New'; font-size:8pt; font-weight:600; 
        font-style:normal; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; 
        -qt-block-indent:0; text-indent:0px; }
</style></head>
<body style="font-family:'Courier New'; font-size:8pt; font-weight:400; font-style:normal">
<br/>
</body></html>'''

        sout = htmlize_diff(sin, 1000)
        self.assertEqual(sout, sout_ref)



#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, Tuple, Any

from PySide2.QtWidgets import QPushButton, QMenu, QAction
from PySide2.QtCore import Signal, QPoint
from PySide2.QtGui import QIcon, QMouseEvent

class MgButtonHistory(QPushButton):
    '''A button to recall history of a text field'''

    historyItemTriggered = Signal(str, str)

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        self.setIcon(QIcon(':img/icons8-history-64.png'))
        self.clicked.connect(self.slotPopupMenu)

        self.popupMenu = QMenu(self)
        self.popupMenu.setToolTipsVisible(True)
        self.popupMenu.triggered.connect(self.slotMenuItemTriggered)
        self.emptyAction = QAction('No items', self)
        self.emptyAction.setEnabled(False)
        self.popupMenu.addAction(self.emptyAction)

        self.clickPoint = QPoint()


    def fillHistory(self, titleList: List[str]) -> None:
        '''Fill the history with a list of items'''
        self.fillHistoryWithTitleAndContent([(v, '') for v in titleList])


    def fillHistoryWithTitleAndContent(self, content: List[Tuple[str, str]]) -> None:
        '''Fill the history with a list of title and content'''
        self.popupMenu.clear()
        for title, message in content:
            action = QAction(title, self)
            action.setToolTip(message)
            self.popupMenu.addAction(action)

        if not self.popupMenu.isEmpty():
            self.emptyAction.setVisible(False)


    def mousePressEvent(self, event: QMouseEvent) -> None:
        # catch event to locate where to put the menu
        self.clickPoint = event.globalPos()
        super().mousePressEvent(event)


    def slotPopupMenu(self) -> None:
        '''Show the menu with the history items'''
        self.popupMenu.popup(self.clickPoint)


    def slotMenuItemTriggered(self, action: QAction) -> None:
        '''One menu item was selected'''
        title = action.text()
        message = action.toolTip()
        self.historyItemTriggered.emit(title, message)





#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


import json
import pathlib, logging
from typing import Dict, Any, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from src.mg_repo_info import MultiRepo, MgRepoInfo

from src.mg_utils import set_username_on_git_url, anonymise_git_url

logger = logging.getLogger('mg_json_mgit_parser')
dbg = logger.debug

# The mgit file format version that we generate
MGIT_FILE_FORMAT_VERSION = "1.0"

# Full format of the mgit file is documented at this page:
# doc/Multigit file format.md


def exportToMgit(mgitFname: str, desc: str, repos: List['MgRepoInfo'], snapshotMode: bool = False) -> None:
    '''Export the list of repositories to a multigit file.

    If snapshotMode is False (the default), the branch/tag information is exported.
    If snapshotMode is True, the tag or SHA1 of the commit of the branch  is exported.
    '''

    repos.sort(key=lambda repo: repo.name.lower())

    reposAsDict = []    # type: List[ Dict[str, str] ]

    for repoInfo in repos:
        # Ensure each repo has the information we need, but normally, this has already been done
        repoInfo.ensure_head_and_url(blocking=True)
        if snapshotMode:
            repoInfo.ensure_sha1(blocking=True)

        # clear username from url
        repoInfo.url = anonymise_git_url(repoInfo.url or '')

        repoDesc = ''
        head: str
        if repoInfo.head and repoInfo.head.find('<empty repo>') == -1:
            if snapshotMode:
                if 'tag' in repoInfo.head:
                    repoDesc = 'commit of %s' % repoInfo.head
                elif 'branch' in repoInfo.head:
                    repoDesc = 'commit on %s' % repoInfo.head

                # force pointing to last commit
                head_type = 'commit'
                assert repoInfo.commit_sha1 is not None # to make mypy happy, we already know that sha1 is set
                head = repoInfo.commit_sha1
            else:
                # branch or tag or commit
                head_type, head = repoInfo.head.split(' ')
        else:
            head_type, head = '', ''

        repoDict = {
            "url": repoInfo.url or '',
            "head": head,
            "head_type": head_type,
            "destination": repoInfo.relpath,
        }
        if len(repoDesc):
            repoDict['description'] = repoDesc

        reposAsDict.append(repoDict)

    reposAsDict.sort(key=lambda rd: rd['destination'].lower())


    mgit_content = {
        'fileFormatVersion': MGIT_FILE_FORMAT_VERSION,
        'description':       desc,
        'repositories':      reposAsDict,

        # for compatibility with previous versions of MultiGit
        'variables' : {},
        'postCloneCommands': [],
     }

    with open(mgitFname, 'w', newline='', encoding='utf8', errors='replace') as jsonf:
        json.dump(mgit_content, jsonf, indent=4, sort_keys=True)



class PostCloneCommand:
    type: str
    description: str
    cmd_file: str

    def __repr__(self) -> str:
        post_cmd_str =  "    POST CLONE COMMAND: {0}\n".format(self.description)
        post_cmd_str += "        type:    {0}\n".format(self.type)
        post_cmd_str += "        file:    {0}\n".format(self.cmd_file)
        post_cmd_str += "\n"
        return post_cmd_str

    def __init__(self, dic_cmd: Dict[str, Any]) -> None:
        self.type = dic_cmd["type"]
        self.description = dic_cmd["description"]
        self.cmd_file = dic_cmd["file"]


class Repository:
    url: str
    head: str
    head_type: str
    destination: str    # relative path to which the clone should be performed
    dest_fullpath: str  # full path of the destination directory
    description: str     # optional description of the head

    def __init__(self, url: str = "", head: str = "", head_type: str = "",
                 dest: str = "", base_path: str = "") -> None:
        self.url = url
        self.head = head
        self.head_type = head_type
        # remove / or \ from the beginning of dest
        self.destination = dest.lstrip("/\\")
        self.update_basepath(base_path)

    def update_basepath(self, base_path: str) -> None:
        '''Propagate the base_path information to all repositories, to fill the full path'''
        self.dest_fullpath = str(pathlib.Path(base_path) / self.destination)

    def fill_from_json(self, dic_rep: Dict[str, str], base_path: str) -> None:
        # JSON input dictionnary format
        #     {
        #         "url": "https://repo_url/proj.git",
        #         "head": "dev",
        #         "head_type": "branch",
        #         "destination": "some/repo/path"
        #     }
        self.url = dic_rep.get("url", "")
        self.head = dic_rep.get("head", "")
        self.head_type = dic_rep.get("head_type", "")
        # remove / or \ from the beginning of destination
        self.destination = dic_rep.get("destination", "").strip("/\\")
        self.description = dic_rep.get("description", "").strip()
        self.update_basepath(base_path)


    def pretty_head(self) -> str:
        '''Pretty name of the head indicating which type of head'''
        s = ' '.join([self.head_type, self.head])
        if len(self.description):
            s += ' (%s)' % self.description
        return s


    def __repr__(self) -> str:
        repo_str =  "    REPOSITORY:\n"
        repo_str += "        URL:    {0}\n".format(self.url)
        if self.description:
            repo_str += "        Description:    {0}\n".format(self.description)
        repo_str += "        Head:    {0}\n".format(self.head)
        if self.head_type:
            repo_str += "        Head type:    {0}\n".format(self.head_type)
        repo_str += "        Destination:    {0}\n".format(self.destination)
        repo_str += "        Full path:    {0}\n".format(self.dest_fullpath)
        repo_str += "\n"
        return repo_str


class ProjectStructure:

    file_format_version: str
    base_path: pathlib.Path
    variables: Dict[str, str]
    description: str
    repos: List['Repository']
    post_commands: List['PostCloneCommand']

    def __init__(self, base_path: str = '') -> None:
        self.base_path = pathlib.Path(base_path)
        self.file_format_version = ''
        self.variables = {}
        self.description = ''
        self.repos = []
        self.post_commands = []

    def __repr__(self) -> str:
        proj_str = "PROJECT:    {0}\n\n".format('\n            '.join(self.description.split('\n')))
        proj_str += "    PATH: {0}\n".format(self.base_path)
        proj_str += '\n'
        if len(self.variables):
            proj_str += "    VARIABLES: \n"
            for k, v in self.variables.items():
                proj_str += "        {0} :    {1}\n".format(k,v)
            proj_str += "\n"
        for repo in self.repos:
            proj_str += "{0}\n".format(repo)
        if len(self.post_commands):
            for post_cmd in self.post_commands:
                proj_str += "{0}\n".format(post_cmd)
        return proj_str


    def replace_variable_in_string(self, input_str: str) -> str:
        '''In the given string, replace the value $VAR$ with the content of the variable VAR.
        Do this for every variable defined'''
        for k, v in self.variables.items():
            if input_str.find("$"+k+"$") != -1:
                input_str = input_str.replace("$"+k+"$", v)
        return input_str


    def apply_variables(self) -> None:
        '''Apply the variable substitution to all fields of the Structure and its Repositories
        and PostClone actions'''
        for k, v in self.variables.items():
            self.variables[k] = self.replace_variable_in_string(v)

        self.description = self.replace_variable_in_string(self.description)

        for repo in self.repos:
            repo.url = self.replace_variable_in_string(repo.url)
            repo.head = self.replace_variable_in_string(repo.head)
            repo.head_type = self.replace_variable_in_string(repo.head_type)
            repo.destination = self.replace_variable_in_string(repo.destination)
            repo.update_basepath(str(self.base_path))

        for post_cmd in self.post_commands:
            post_cmd.description = self.replace_variable_in_string(post_cmd.description)
            post_cmd.type = self.replace_variable_in_string(post_cmd.type)
            post_cmd.cmd_file = self.replace_variable_in_string(post_cmd.cmd_file)


    def fill_repositories_from_json_data(self, repo_list: List[Dict[str, str]]) -> None:
        self.repos = []
        for dic_rep in repo_list:
            rep = Repository()
            rep.fill_from_json(dic_rep, str(self.base_path))
            self.repos.append(rep)
        self.repos.sort(key=lambda repo: repo.destination.lower())


    def fill_postcommand_from_json_data(self, post_cmd_list: List[Dict[str, Any]]) -> List['PostCloneCommand']:
        self.post_commands = []
        for dic_cmd in post_cmd_list:
            self.post_commands.append(PostCloneCommand(dic_cmd))
        return self.post_commands


    def fill_from_json_data(self, d: Dict[str, Any]) -> None:
        self.variables = d.get("variables", {})
        self.description = d["description"]
        self.file_format_version = d["fileFormatVersion"]
        self.fill_repositories_from_json_data(d["repositories"])
        self.fill_postcommand_from_json_data(d.get("postCloneCommands", []))
        self.apply_variables()


    def set_base_path(self, base_path: pathlib.Path) -> None:
        self.base_path = base_path
        for repo in self.repos:
            repo.update_basepath(str(self.base_path))


    def fill_from_json_file(self, project_json_file: Union[pathlib.Path, str],
                            base_path: pathlib.Path) -> None:
        '''Parse the json file. May generate json.JSONDecodeError if invalid json file is provided'''
        project_json_file = pathlib.Path(project_json_file)
        with open(project_json_file, encoding='utf8') as file:
            data = json.load(file)
        self.set_base_path(base_path)
        self.fill_from_json_data(data)




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import cast, List, Optional, Union
import logging, pathlib

from PySide2.QtWidgets import QMessageBox, QApplication, QWidget
from PySide2.QtCore import Qt, Signal, QByteArray, QTimer

from src.mg_repo_info import MultiRepo
from src.mg_repo_tree_item import MgRepoTreeItem
from src.mg_utils import htmlize_diff
from src import mg_config as mgc
from src.mg_const import COL_REPO_NAME, MAX_DIFF_LINES
from src.gui.ui_multigit_widget import Ui_MultigitWidget

logger = logging.getLogger('mg_multigit_widget')
dbg = logger.debug
info = logger.info


class MgMultigitWidget(QWidget, Ui_MultigitWidget):

    multiRepo: MultiRepo

    sig_request_dir_open = Signal()
    sig_dir_changed = Signal(str)

    def __init__(self) -> None:
        super().__init__()
        self.setupUi(self)

        self.multiRepo = MultiRepo('')

        self.config = mgc.get_config_instance()

        # basedir buttons
        self.buttonHistoryBaseDir.historyItemTriggered.connect( lambda title, message: self.openDir(title) )
        self.buttonHistoryBaseDir.fillHistory(self.config.lruAsList(mgc.CONFIG_LAST_OPENED))
        self.buttonOpenBaseDir.clicked.connect(self.sig_request_dir_open)


        if self.config[mgc.CONFIG_SPLITTER_STATE_V2]:
            state: List[int] =  self.config[mgc.CONFIG_SPLITTER_STATE_V2]
            bstate = QByteArray(bytes(state))
            self.splitter.restoreState(bstate)
        elif self.config[mgc.CONFIG_SPLITTER_STATE]:
            self.splitter.restoreState(self.config[mgc.CONFIG_SPLITTER_STATE_V2])

        self.repoTree.itemSelectionChanged.connect(self.slotItemSelectionChanged)

        self.tabWidget.setCurrentIndex(0)
        self.tabWidget.currentChanged.connect( self.slotTabWidgetIndexChanged )


##########################################################################
#
#       Methods related to the display updates and internal machinery
#
##########################################################################


    def setBaseDir(self, baseDir: Optional[str]) -> None:
        '''Set the base directory from which all git repos are searched.'''
        dbg('setBaseDir( %s )' % str(baseDir) )

        if baseDir and not pathlib.Path(baseDir).exists():
            QMessageBox.warning(self, 'Could not open path', 'Error: could not open path: %s' % baseDir)
            return

        self.sig_dir_changed.emit( baseDir or '' )
        self.repoTree.clear()
        if baseDir is None:
            self.lineEditBaseDir.setEnabled(False)
            self.lineEditBaseDir.setText("")
            self.multiRepo = MultiRepo('')
            return

        self.lineEditBaseDir.setEnabled(True)
        self.lineEditBaseDir.setText(baseDir)
        self.multiRepo = MultiRepo(baseDir)
        QTimer.singleShot(0, self.fillRepoView)


    def fillRepoView(self) -> None:
        '''Fill the repoTree with the repository information'''
        dbg('fillRepoView()')
        if self.multiRepo.isEmpty():
            # very first launch of MultiGit without base dir defined
            # or selection of a directory with no git repo
            self.repoTree.clear()
            return

        self.repoTree.setFocus(Qt.OtherFocusReason)
        try:
            self.setCursor(Qt.WaitCursor)
            QApplication.processEvents()
            # we keep that one blocking, it's very quick anyways
            self.multiRepo.find_git_repos()

            items = self.repoTree.addRepos(self.multiRepo.repo_list)

        finally:
            # all items are visible now, we can restore the cursor
            self.setCursor(Qt.ArrowCursor)

        # 2nd step, we fill the repo information
        for it in items:
            it.repoInfo.refresh()


    def slotItemSelectionChanged(self) -> None:
        items = self.selectedItems()
        dbg('slotItemSelectionChanged() - %d items, %s' % (len(items), items[0].text(COL_REPO_NAME) if len(items) else ''))

        if not self.tabWidget.isVisible():
            # if tab widget is not visible, we don't need to update its content when current item changes
            return

        if len(items) == 0:
            # nothing is selected
            self.textEditModFiles.setText('')
            self.textEditCommit.setText('')
            return

        # force refresh of the correct tab item
        self.slotTabWidgetIndexChanged(self.tabWidget.currentIndex())


    def slotTabWidgetIndexChanged(self, idx: int) -> None:
        '''Called when the tab widget between diff and modified files is changed'''
        dbg('slotTabWidgetIndexChanged(%d)' % idx)
        items = self.selectedItems()
        if len(items) == 0:
            return

        currentRepo = items[0].repoInfo
        if self.tabWidget.widget(idx) == self.tabLastCommit:
            # showing last commit currently
            currentRepo.ensure_last_commit(self.slotSetLastCommit)
        else:
            # showing diff currently
            currentRepo.ensure_diff(self.displayRepoDiff)


    def slotSetLastCommit(self, repo_name: str, last_commit: str) -> None:
        dbg('slotSetLastCommit("%s", ...)' % repo_name)
        currentItem = self.repoTree.currentItem()
        if currentItem is None:
            return

        currentRepoName = currentItem.text(COL_REPO_NAME)
        if repo_name != currentRepoName:
            logger.warning('Received last commit result for repo "%s" but repo selected is "%s". Call ignored' % (repo_name, currentRepoName))
            # the selection changed between the time where the lastCommit info was requested and it was received
            # just ignore it
            return

        self.textEditCommit.setText(last_commit)


    def displayRepoDiff(self, repo_name: str, repo_diff: str) -> None:
        dbg('displayRepoDiff("%s", ...)' % repo_name)
        currentItem = self.repoTree.currentItem()
        if currentItem is None:
            return

        currentRepoName = currentItem.text(COL_REPO_NAME)
        if repo_name != currentRepoName:
            logger.warning('Received diff result for repo "%s" but repo selected is "%s". Call ignored' % (repo_name, currentRepoName))
            # the selection changed between the time where the lastCommit info was requested and it was received
            # just ignore it
            return

        self.textEditModFiles.setText(htmlize_diff(repo_diff or '', MAX_DIFF_LINES))


    def slotViewTabChanged(self, showTabLastCommit: bool, showTabModFiles: bool) -> None:
        '''Called when the menu Edit->View->Tab Last Commit or View->Tab Modified Files changes'''
        dbg('slotViewTabChanged')
        if (not showTabLastCommit) and (not showTabModFiles):
            # nothing to show in the tab, hide it
            self.tabWidget.setVisible(False)
        else:
            self.tabWidget.setVisible(True)
            showHideTabLastCommitFuncDict = {
                (False, False): lambda: None, # hidden and requested hidden, do nothing
                (False, True):  lambda: self.tabWidget.insertTab(0, self.tabLastCommit, 'Last Commit'),  # hidden and requested visible, show it
                (True, False):  lambda: self.tabWidget.removeTab(self.tabWidget.indexOf(self.tabLastCommit)),
                (True, True) :  lambda: None, # visible and requested visible, do nothing
            }
            showHideTabLastCommitFuncDict[(self.tabWidget.indexOf(self.tabLastCommit) != -1, showTabLastCommit)]()


            showHideTabModFilesFuncDict = {
                (False, False): lambda: None, # hidden and requested hidden, do nothing
                (False, True):  lambda: self.tabWidget.insertTab(1, self.tabModifiedFiles, 'Modified Files'),  # hidden and requested visible, show it
                (True, False):  lambda: self.tabWidget.removeTab(self.tabWidget.indexOf(self.tabModifiedFiles)),
                (True, True) :  lambda: None, # visible and requested visible, do nothing
            }
            showHideTabModFilesFuncDict[(self.tabWidget.indexOf(self.tabModifiedFiles) != -1, showTabModFiles)]()

        # trigger refresh of the tab widget if it was hidden
        self.slotTabWidgetIndexChanged(self.tabWidget.currentIndex())


##########################################################################
#
#       Methods related to menu entries NOT acting on selected items
#
##########################################################################

    def openDir(self, newDir: str, ignoreNonExisting: bool = False) -> None:
        info('Opening dir: %s' % newDir)
        self.buttonHistoryBaseDir.fillHistory(self.config.lruAsList(mgc.CONFIG_LAST_OPENED))
        self.setBaseDir(newDir)


    def slotRefreshAll(self) -> None:
        dbg('refreshAll()')
        self.repoTree.setFocus(Qt.OtherFocusReason)
        try:
            self.setCursor(Qt.WaitCursor)
            QApplication.processEvents()
            # we keep that one blocking, it's very quick anyways
            added_repo, rm_repo = self.multiRepo.find_git_repos_added_removed()

            self.repoTree.deleteRepos(rm_repo)

            # Add items related to new repositories
            self.repoTree.addRepos(added_repo)

        finally:
            self.setCursor(Qt.ArrowCursor)
            QApplication.processEvents()

        # 2nd step, update multi-repo
        self.multiRepo.adjust_git_repos(added_repo, rm_repo)

        # 3rd step: update all the repos
        for it in self.multiRepo.repo_list:
            it.refresh()


    def slotGitFetchAll(self) -> None:
        '''Run git fetch on all repos'''
        dbg('slotGitFetchAll()')
        # collect the items in the order they are shown
        items = [cast(MgRepoTreeItem, self.repoTree.topLevelItem(idx))
                 for idx in range(0, self.repoTree.topLevelItemCount())]
        repos = [item.repoInfo for item in items]

        cmd = ['fetch', '--prune']
        self.repoTree.getGitExecWindow().execOneGitCommand('Fetching all repositories', cmd, repos)


    def selectedItems(self) -> List[MgRepoTreeItem]:
        '''Return the list of selected MgRepoTreeItem'''
        items = cast(List[MgRepoTreeItem], list(self.repoTree.selectedItems()))
        return items


